import pandas as pd
import numpy as np
from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook
from sklearn.base import clone
import time
import warnings
from ..dataset.helpers import verbose_print
from ..pipeline.helpers import get_mean_fis
from sklearn.utils import Bunch
from scipy.stats import t
from .stats_helpers import corrected_std, compute_corrected_ttest


def is_notebook():

    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
    except NameError:
        pass
    return False


# @TODO
# 1. Store permutation FI's in object after call
# 2. Add methods for plot feature importance's ?


def get_non_nan_Xy(X, y):

    # Check for if any missing targets in the test set to skip
    if pd.isnull(y).any():
        non_nan_subjs = y[~pd.isnull(y)].index
        X, y = X.loc[non_nan_subjs], y.loc[non_nan_subjs]

    return X, y


def fi_to_series(fi, feat_names):

    try:
        fi = fi.squeeze()
    except AttributeError:
        pass

    # Base flat case
    if len(fi.shape) == 1:
        return pd.Series(fi, index=feat_names)

    # Categorical case
    # @TODO is there a better way to have this?
    # e.g., maybe explicitly by class value, see self.estimators[0].classes_
    # could put it as another index level.

    series = []
    for class_fi in fi:
        series.append(pd.Series(class_fi, index=feat_names))

    return series


def fis_to_df(fis):

    # Base case - assume that first element is representative
    if isinstance(fis[0], pd.Series):
        return pd.DataFrame(fis)

    # Categorical case
    dfs = []
    for c in range(len(fis[0])):
        dfs.append(pd.DataFrame([fi[c] for fi in fis]))

    return dfs


def mean_no_zeros(df):

    mean = df.mean()
    return mean[mean != 0]


class BPtEvaluator():
    '''This class is returned from calls to :func:`evaluate`,
    and can be used to store information from evaluate, or
    compute additional feature importances.'''

    # Add verbose print
    _print = verbose_print

    def __init__(self, estimator, ps,
                 encoders=None,
                 progress_bar=True,
                 store_preds=False,
                 store_estimators=False,
                 store_timing=False,
                 eval_verbose=0,
                 progress_loc=None,
                 compare_bars=None):

        # Save base
        self.estimator = estimator
        self.ps = ps
        self.encoders_ = encoders

        # Set if using progress bar
        self._set_progress_bar(progress_bar)

        # If store preds
        self.preds = None
        if store_preds:
            self.preds = {}

        # If store estimator
        self.estimators = None
        if store_estimators:
            self.estimators = []

        # If store timing
        self.timing = None
        if store_timing:
            self.timing = {'fit': [], 'score': []}

        # @TODO Add in progress loc func.
        self.progress_loc = progress_loc
        self.verbose = eval_verbose
        self.compare_bars = compare_bars

    @property
    def mean_scores(self):
        '''This parameter stores the mean scores as
        a dictionary of values, where each dictionary
        is indexed by the name of the scorer, and the dictionary value
        is the mean score for that scorer.'''
        return self._mean_scores

    @mean_scores.setter
    def mean_scores(self, mean_scores):
        self._mean_scores = mean_scores

    @property
    def std_scores(self):
        '''This parameter stores the standard deviation scores as
        a dictionary of values, where each dictionary
        is indexed by the name of the scorer, and value
        contains the standard deviation across evaluation folds
        for that scorer.

        The default scorer key stores the micro standard
        deviation, but in the case that macro standard deviation differs,
        i.e., in the case of multiple repeats in an evaluation, then
        a separate macro standard deviation will be stored under
        the name of the scorer with _macro appended to the key.

        For example if a 3-fold twice repeated evaluation was
        run, with just r2 as the scorer, this parameter might
        look something like:

        ::

            self.std_scores = {'r2': .5, 'r2_macro': .01}

        '''
        return self._std_scores

    @std_scores.setter
    def std_scores(self, std_scores):
        self._std_scores = std_scores

    @property
    def weighted_mean_scores(self):
        '''This property stores the mean scores
        across evaluation folds (simmilar to
        :data:`mean_scores<BPtEvaluator.mean_scores>`),
        but weighted by the
        number of subjects / datapoints in each fold.

        It is scores as a dictionary indexed by the name
        of the scorer as the key, where values are
        the weighted mean score.
        '''
        return self._weighted_mean_scores

    @weighted_mean_scores.setter
    def weighted_mean_scores(self, weighted_mean_scores):
        self._weighted_mean_scores = weighted_mean_scores

    @property
    def scores(self):
        '''This property stores the scores for
        each scorer as a dictionary of lists, where
        the keys are the names of the scorer and the list
        represents the score obtained for each fold, where each
        index corresponds to to a fold of cross validation.'''
        return self._scores

    @scores.setter
    def scores(self, scores):
        self._scores = scores

    @property
    def score(self):
        '''This property represents
        a quick helper for accessing the mean scores
        of whatever the first scorer is (in the case of
        multiple scorers).
        '''

        first_scorer = list(self.mean_scores)[0]
        return self.mean_scores[first_scorer]

    @property
    def ps(self):
        '''A saved and pre-processed version of the problem_spec
        used (with any extra_params applied) when running this
        instance of Evaluator.'''
        return self._ps

    @ps.setter
    def ps(self, ps):
        self._ps = ps

    @property
    def feat_names(self):
        '''The features names corresponding to any measures of
        feature importance, stored as a list of lists, where the top
        level list represents each fold of cross validation.

        This parameter may be especially useful when pipeline
        objects such as transformers or feature selectors are used
        as these can drastically change the features passed to an
        eventual model.

        The values stored here may change
        based on the passed
        value of the `decode_feat_names` parameter from
        :func:`evaluate`.

        For example the feat_names from a 3-fold cross-validation
        with input features ['feat1', 'feat2', 'feat3'] with
        feature selection as a piece of the pipeline may look like:

        ::

            self.feat_names = [['feat1', 'feat2'],
                               ['feat2', 'feat3'],
                               ['feat1', 'feat2']]

        '''
        return self._feat_names

    @feat_names.setter
    def feat_names(self, feat_names):
        self._feat_names = feat_names

    @property
    def val_indices(self):
        '''| This parameter stores the validation subjects / index
          used in every fold of the cross-validation. It can be
          useful in some cases
          to check to see exactly what cross-validation was applied.

        | This parameter
          differs from
          :data:`all_val_indices<BPtEvaluator.all_val_indices>`
          in that even subjects with missing target values are not included.

        '''
        return self._val_indices

    @val_indices.setter
    def val_indices(self, val_indices):
        self._val_indices = val_indices

    @property
    def train_indices(self):
        '''| This parameter stores the training subjects / index
          used in every fold of the cross-validation. It can be
          useful in some cases to check to see exactly what
          cross-validation was applied.

        | This parameter
          differs from
          :data:`all_train_indices<BPtEvaluator.all_train_indices>`
          in that even subjects with missing target values are not included.

        '''
        return self._train_indices

    @train_indices.setter
    def train_indices(self, train_indices):
        self._train_indices = train_indices

    @property
    def all_val_indices(self):
        '''| This parameter stores the validation subjects / index
          used in every fold of the cross-validation.

        | This parameter
          differs from :data:`val_indices<BPtEvaluator.val_indices>`
          in that even subjects with missing target values are included.

        '''
        return self._all_val_indices

    @all_val_indices.setter
    def all_val_indices(self, all_val_indices):
        self._all_val_indices = all_val_indices

    @property
    def all_train_indices(self):
        '''| This parameter stores the training subjects / index
          used in every fold of the cross-validation.

        | This parameter
          differs from :data:`train_indices<BPtEvaluator.train_indices>`
          in that even subjects with missing target values are included.
        '''
        return self._all_train_indices

    @all_train_indices.setter
    def all_train_indices(self, all_train_indices):
        self._all_train_indices = all_train_indices

    @property
    def timing(self):
        '''This property stores information on
        the fit and scoring times, if requested by the
        original call to :func:`evaluate`.
        This parameter is a dictionary with two keys,
        'fit' and 'score'.
        Each key stores the time in seconds as a list of
        values for each of the evaluation folds.
        '''
        return self._timing

    @timing.setter
    def timing(self, timing):
        self._timing = timing

    @property
    def mean_timing(self):
        '''This property stores information on
        the fit and scoring times, if requested by the
        original call to :func:`evaluate`.
        This parameter is a dictionary with two keys,
        'fit' and 'score'.
        Each key stores the mean time in seconds across folds.
        '''
        return self._mean_timing

    @mean_timing.setter
    def mean_timing(self, mean_timing):
        self._mean_timing = mean_timing

    @property
    def preds(self):
        '''If the parameter `store_preds` is set to True when
        calling :func:`evaluate`, then this parameter will store the
        predictions from every evaluate fold.

        The parameter preds is a dictionary, where raw predictions made
        can be accessed by the key 'predict'. Values are stored as list
        corresponding to each evaluation fold.

        In the case where other predict-like functions are avaliable, e.g.,
        in the case of a binary problem, where it may be desirable to
        see the predicted probability, then the those predictions
        will be made avaliable under the name of the underlying predict
        function. In this case, that is self.preds['predict_proba'].
        It will also store results from 'predict' as well.

        self.preds also will store under 'y_true' a list, where
        each element of the list corresponds to the corresponding
        true target values for the predictions made.
        '''

        return self._preds

    @preds.setter
    def preds(self, preds):
        self._preds = preds

    @property
    def estimators(self):
        '''If the parameter `store_estimators` is set to True when
        calling :func:`evaluate`, then this parameter will store the fitted
        estimator in a list. Where each element of the list corresponds to one
        of the validation folds.

        For example to access the fitted estimator from this first
        fold ::

            first_est = self.estimators[0]

        '''
        return self._estimators

    @estimators.setter
    def estimators(self, estimators):
        self._estimators = estimators

    def _set_progress_bar(self, progress_bar):

        if not progress_bar:
            self.progress_bar = None
        elif is_notebook():
            self.progress_bar = tqdm_notebook
        else:
            self.progress_bar = tqdm

    def _eval(self, X, y, cv):

        # If verbose is lower than -1,
        # then don't show any warnings no matter the source.
        if self.verbose < -1:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._evaluate(X, y, cv)

        # Otherwise, base behavior
        else:
            self._evaluate(X, y, cv)

    def _evaluate(self, X, y, cv):
        '''cv is passed as raw index, X and y as dataframes.'''

        # Compute and warn about num nan targets
        n_nan_targets = pd.isnull(y).sum()
        if n_nan_targets > 0:
            self._print('Warning: There are', str(n_nan_targets) + ' missing',
                        'targets passed to evaluate. Subjects with missing',
                        'target values will be skipped during training and '
                        'scoring.')

            if self.preds is not None:
                self._print('Note: Predictions will still be made for any',
                            'subjects with missing values in ',
                            'any validation folds.')

        # Verbose info
        self._print('Predicting target =', str(self.ps.target), level=1)
        self._print('Using problem_type =', str(self.ps.problem_type), level=1)
        self._print('Using scope =', str(self.ps.scope),
                    '(defining a total of', str(X.shape[1]),
                    'features).', level=1)
        self._print(f'Evaluating {len(X)} total data points.', level=1)

        # Init scores as dictionary of lists
        self.scores = {scorer_str: [] for scorer_str in self.ps.scorer}

        # Save train and test subjs
        self.all_train_indices, self.all_val_indices = [], []
        self.train_indices, self.val_indices = [], []

        # Save final feat names
        self.feat_names = []

        # Init progress bar / save and compute fold info from cv
        progress_bars = self._init_progress_bars(cv)

        self._print('Using CV: ', cv,
                    'to generate evaluation splits.', level=2)
        self._print(level=1)

        # Run each split
        for train_inds, val_inds in cv.split(X, y):

            # Eval
            self._eval_fold(X.iloc[train_inds], y.iloc[train_inds],
                            X.iloc[val_inds], y.iloc[val_inds])

            # Increment progress bars
            progress_bars = self._incr_progress_bars(progress_bars)

        # Clean up progress bars
        self._finish_progress_bars(progress_bars)

        # Compute and score mean and stds
        self._compute_summary_scores()

    def _init_progress_bars(self, cv):

        # Passed cv should have n_repeats param - save in class
        self.n_repeats_ = 1
        if hasattr(cv, 'n_repeats'):
            self.n_repeats_ = cv.n_repeats

        # Passed cv should already be sklearn style
        n_all_splits = cv.get_n_splits()

        # Compute number of splits per repeat
        self.n_splits_ = n_all_splits
        if self.n_repeats_ > 1:
            self.n_splits_ = int(n_all_splits / self.n_repeats_)

        # Skip if no progress bar
        if self.progress_bar is None:
            return []

        # If passed compare bars is int, init top level bar
        if isinstance(self.compare_bars, int):

            # Init and set as new
            compare_bar = self.progress_bar(total=self.compare_bars,
                                            desc='Compare')
            self.compare_bars = [compare_bar]

        # If already init'ed
        elif isinstance(self.compare_bars, list):

            # Return all but last compare bar
            return self.compare_bars[:-1]

        bars = []

        # If 1 repeat, then just folds progress bar
        if self.n_repeats_ == 1:
            folds_bar = self.progress_bar(total=self.n_splits_, desc='Folds')
            bars = [folds_bar]

        # Otherwise folds and repeats bars - init repeats bar first, so on top
        else:
            repeats_bar = self.progress_bar(total=self.n_repeats_,
                                            desc='Repeats')
            folds_bar = self.progress_bar(total=self.n_splits_, desc='Folds')
            bars = [folds_bar, repeats_bar]

        # If compare bars was init'ed this run
        if self.compare_bars is not None:
            self.compare_bars = bars + self.compare_bars

        return bars

    def _incr_progress_bars(self, progress_bars):

        # Skip if not requested
        if self.progress_bar is None:
            return []

        # Increment folds bar
        folds_bar = progress_bars[0]
        folds_bar.n += 1

        # If just folds bar update and return
        if len(progress_bars) == 1:
            folds_bar.refresh()
            return [folds_bar]

        # If both, check to see if n_repeats increments
        repeats_bar = progress_bars[1]
        if folds_bar.n == self.n_splits_:
            folds_bar.n = 0
            repeats_bar.n += 1

            # If end, set to full
            if repeats_bar.n == self.n_repeats_:
                folds_bar.n = self.n_splits_

        # Update and return
        folds_bar.refresh()
        repeats_bar.refresh()
        return [folds_bar, repeats_bar]

    def _finish_progress_bars(self, progress_bars):

        # Close progress bars
        if self.compare_bars is None:
            for bar in progress_bars:
                bar.close()

            return

        # Otherwise compare bars case
        # Reset
        for bar in progress_bars:
            bar.n = 0
            bar.refresh()

        # Increment and refresh compare
        self.compare_bars[-1].n += 1
        self.compare_bars[-1].refresh()

        return

    def _eval_fold(self, X_tr, y_tr, X_val, y_val):

        # Get clone of estimator to fit
        estimator_ = clone(self.estimator)

        # Save all train and val inds before missing targets removed
        self.all_train_indices.append(X_tr.index)
        self.all_val_indices.append(X_val.index)

        # Check for if any missing targets, if so - skip
        # those subjects.
        X_tr, y_tr = get_non_nan_Xy(X_tr, y_tr)
        X_val_c, y_val_c = get_non_nan_Xy(X_val, y_val)

        # Keep track of subjects in folds - where a subject is not included
        # in the train or val fold if has NaN target
        self.train_indices.append(X_tr.index)
        self.val_indices.append(X_val_c.index)

        self._print('Train size:', len(X_tr), '- Val size:',
                    len(X_val_c), level=1)

        # Print if skipping any due to NaN target
        dif_tr = len(self.all_train_indices[-1]) - len(self.train_indices[-1])
        dif_val = len(self.all_train_indices[-1]) - len(self.train_indices[-1])
        if dif_tr != 0 or dif_val != 0:
            self._print(f'Skipping Train: {dif_tr} - Val: {dif_val},',
                        'for NaN target values.', level=1)

        # Fit estimator_, passing as arrays, and with train data index
        start_time = time.time()

        estimator_.fit(X=X_tr, y=np.array(y_tr))
        fit_time = time.time() - start_time
        self._print(f'Fit fold in {fit_time:.3f} seconds.', level=1)

        # Score estimator
        start_time = time.time()
        self._score_estimator(estimator_, X_val_c, y_val_c)
        score_time = time.time() - start_time

        # Store timing if requested
        if self.timing is not None:
            self.timing['fit'].append(fit_time)
            self.timing['score'].append(score_time)

        # Save preds - pass full val with NaN targets
        self._save_preds(estimator_, X_val, y_val)

        # Get and save final transformed feat names
        self.feat_names.append(
            estimator_.transform_feat_names(X_tr,
                                            encoders=self.encoders_))

        # If store estimators, save in self.estimators
        if self.estimators is not None:
            self.estimators.append(estimator_)

    def _score_estimator(self, estimator_, X_val, y_val):

        # Save score for each scorer
        for scorer_str in self.ps.scorer:
            score = self.ps.scorer[scorer_str](estimator_,
                                               X_val,
                                               np.array(y_val))
            self.scores[scorer_str].append(score)
            self._print(scorer_str + ':', str(score), level=1)

    def _save_preds(self, estimator, X_val, y_val):

        if self.preds is None:
            return

        self._print('Saving predictions on validation set.', level=2)

        for predict_func in ['predict', 'predict_proba', 'decision_function']:

            # Get preds, skip if estimator doesn't have predict func
            try:
                preds = getattr(estimator, predict_func)(X_val)
            except AttributeError:
                continue

            # Add to preds dict if estimator has predict func
            try:
                self.preds[predict_func].append(preds)
            except KeyError:
                self.preds[predict_func] = [preds]

        # Add y_true
        try:
            self.preds['y_true'].append(np.array(y_val))
        except KeyError:
            self.preds['y_true'] = [np.array(y_val)]

    def _compute_summary_scores(self):

        self._print('Computing summary scores.', level=2)

        self.mean_scores, self.std_scores = {}, {}
        self.weighted_mean_scores = {}

        for scorer_key in self.scores:

            # Save mean under same name
            scores = self.scores[scorer_key]
            self.mean_scores[scorer_key] = np.mean(scores)

            # Compute scores weighted by number of subjs
            # Use val_indices without NaN targets
            weights = [len(self.val_indices[i])
                       for i in range(len(self.val_indices))]
            self.weighted_mean_scores[scorer_key] =\
                np.average(scores, weights=weights)

            # Compute and add base micro std
            self.std_scores[scorer_key] = np.std(scores)

            # If more than 1 repeat, add the macro std
            if self.n_repeats_ > 1:
                scores = np.reshape(scores,
                                    (self.n_repeats_, self.n_splits_))
                self.std_scores[scorer_key + '_macro'] =\
                    np.std(np.mean(scores, axis=1))

        # Add mean timing
        if self.timing is not None:
            self.mean_timing = {}

            for time_key in self.timing:
                self.mean_timing[time_key] = np.mean(self.timing[time_key])

    def get_preds_dfs(self, drop_nan_targets=False):
        '''This function can be used to return the raw predictions
        made during evaluation as a list of pandas Dataframes.

        Parameters
        ------------
        drop_nan_targets : bool, optional
            If False (default), then this method will return the
            DataFrame of predictions including targets
            with NaN. To skip these, e.g., in this case
            of plotting against ground truth or computing
            new metrics, set to True.

            ::

                default = False

        Returns
        ---------
        dfs : list of pandas.DataFrame
            list of dataframe's per fold, where each DataFrame
            contains predictions made.
        '''

        dfs = []

        # For each fold
        for fold_indx in range(len(self.all_val_indices)):

            # Init df
            df = pd.DataFrame(index=self.all_val_indices[fold_indx])

            # Add each predict type as a column
            for predict_type in self.preds:
                ps = self.preds[predict_type][fold_indx]

                # Either float or multi-class case
                if isinstance(ps[0], float):
                    df[predict_type] = ps

                else:
                    for cls in range(len(ps[0])):
                        df[predict_type + '_' + str(cls)] = ps[:, cls]

            # Drop nan-cols if not requested
            if drop_nan_targets:
                nan_targets = df[df['y_true'].isna()].index
                df = df.drop(nan_targets)

            # Add to by fold list
            dfs.append(df)

        return dfs

    def __repr__(self):
        rep = 'BPtEvaluator\n'
        rep += '------------\n'

        # Add scores + means
        rep += 'mean_scores = ' + repr(self.mean_scores) + '\n'
        rep += 'std_scores = ' + repr(self.std_scores) + '\n'
        rep += '\n'

        # Show avaliable saved attrs
        saved_attrs = []
        avaliable_methods = []

        if self.estimators is not None:
            saved_attrs.append('estimators')
        if self.preds is not None:
            saved_attrs.append('preds')
            avaliable_methods.append('get_preds_dfs')
        if self.timing is not None:
            saved_attrs.append('timing')

        saved_attrs += ['train_indices', 'val_indices', 'feat_names', 'ps',
                        'mean_scores', 'std_scores',
                        'weighted_mean_scores', 'scores']

        # Only show if different
        ati_len = len(sum([list(e) for e in self.all_train_indices], []))
        ti_len = len(sum([list(e) for e in self.train_indices], []))
        if ati_len != ti_len:
            saved_attrs.append('all_train_indices')

        avi_len = len(sum([list(e) for e in self.all_val_indices], []))
        vi_len = len(sum([list(e) for e in self.val_indices], []))
        if avi_len != vi_len:
            saved_attrs.append('all_val_indices')

        if self.estimators is not None:

            # Either or
            if self.feature_importances_ is not None:
                saved_attrs += ['fis_', 'feature_importances_']
                avaliable_methods += ['get_fis', 'get_feature_importances']
            elif self.coef_ is not None:
                saved_attrs += ['fis_', 'coef_']
                avaliable_methods += ['get_fis', 'get_coef_']

            avaliable_methods.append('permutation_importance')

        rep += 'Saved Attributes: ' + repr(saved_attrs) + '\n\n'
        rep += 'Avaliable Methods: ' + repr(avaliable_methods) + '\n\n'
        rep += 'Evaluated with:\n' + repr(self.ps) + '\n'

        return rep

    def _repr_html_(self):
        pass

    def _estimators_check(self):

        if self.estimators is None:
            raise RuntimeError('This method is not avaliable unless '
                               'evaluate is run with store_estimators=True!')

    @property
    def feature_importances_(self):
        '''This property stores the mean values
        across fitted estimators assuming each fitted estimator
        has a non empty `feature_importances_` attribute.'''

        self._estimators_check()
        return get_mean_fis(self.estimators, 'feature_importances_')

    def get_feature_importances(self):
        '''This function returns each `feature_importances_`
        value across fitted estimators. If None have this parameter,
        it will return a list of None.

        Returns
        --------
        feature_importances : list
            A list of `feature_importances_` where each element
            in the list refers to a fold from the evaluation.
        '''

        self._estimators_check()
        return [estimator.feature_importances_
                for estimator in self.estimators]

    @property
    def coef_(self):
        '''This attribute represents the mean `coef_` as
        a numpy array across all folds. This parameter will only
        be avaliable if all estimators have a non null `coef_` parameter
        and each returns the same shape. See `fis_` for a more flexible
        version of this parameter that can handle when there
        are differing numbers of features.'''

        self._estimators_check()
        return get_mean_fis(self.estimators, 'coef_')

    def get_coefs(self):
        '''This function returns each `coef_`
        value across fitted estimators. If None have this parameter,
        it will return a list of None.

        Returns
        --------
        coefs : list
            A list of `coef_` where each element
            in the list refers to a fold from the evaluation.
        '''

        self._estimators_check()
        return [estimator.coef_
                for estimator in self.estimators]

    @property
    def fis_(self):
        '''This property stores the mean value
        across each fold of the CV for either the `coef_`
        or `feature_importance_` parameter.

        Warnings
        ---------
        If a feature is not present in all folds,
        then it's mean value will be computed from only the
        folds in which it was present.

        When using transformers, for example one hot encoder,
        since the encoding is done on the fly, there is no
        guarantee that 'one hot encoder category_1' is actually
        the same category 1 across folds.

        If for some reason some folds have a model with feature
        importances and other `coef_` they will still all be averaged
        together, so make sure that this parameter is only used when
        all of the underlying models across folds should have comparable
        feature importances.
        '''

        # @TODO incoperate in information about the original
        # class names here // maybe in specific objects like
        # OneHotEncoder.

        self._estimators_check()

        # Grab fis as Dataframe or list of
        fis = self.get_fis()

        # Base case
        if isinstance(fis, pd.DataFrame):
            return fis.mean()

        # Categorical case
        return [fi.mean() for fi in fis]

    def _get_base_fis_list(self):

        self._estimators_check()

        coefs = self.get_coefs()
        feature_importances = self.get_feature_importances()

        fis = []
        for coef, fi, feat_names in zip(coefs, feature_importances,
                                        self.feat_names):
            if coef is not None:
                fis.append(fi_to_series(coef, feat_names))
            elif fi is not None:
                fis.append(fi_to_series(fi, feat_names))
            else:
                fis.appends(None)

        return fis

    def get_fis(self, mean=False, abs=False):
        '''This method will return a pandas DataFrame with
        each row a fold, and each column a feature if
        the underlying model supported either the `coef_`
        or `feature_importance_` parameters.

        In the case that the underlying feature importances
        or `coefs_` are not flat, e.g., in the case
        of a one versus rest categorical model, then a list
        multiple DataFrames will be returned, one for each class.
        The order of the list will correspond to the order of classes.

        Parameters
        -----------
        mean : bool, optional
            If True, return the mean value
            across evaluation folds as a pandas Series.
            Any features with a mean value of 0 will
            also be excluded. Otherwise, if default
            of False, return raw values for each fold
            as a Dataframe.

            ::

                default = False

        abs : bool, optional
            If the feature importances
            should be absolute values
            or not.

            ::

                default = False

        Returns
        --------
        fis : pandas DataFrame or Series
            Assuming mean=False, the
            a pandas DataFrame where each row contains the
            feature importances from an evaluation fold (unless the underlying
            feature importances are categorical, in which a list of DataFrames
            will be returned.)

            If mean=True, then a pandas Series (or in the case of
            underlying categorical feature importances, list of)
            will be returned, with the mean value from each fold
            and all features with a value of 0 excluded.

            Note: To get the mean values without zero's excluded,
            just call .mean() on the result of this method
            with mean=False.


        '''

        fis = self._get_base_fis_list()
        base = fis_to_df(fis)

        # Proc. abs arg
        if abs:
            if isinstance(base, list):
                base = [b.abs() for b in base]
            else:
                base = base.abs()

        # If not mean, return as is
        if not mean:
            return base

        # Categorical mean case
        if isinstance(base, list):
            return [mean_no_zeros(b) for b in base]

        # Base mean case
        return mean_no_zeros(base)

    def get_inverse_fis(self):
        '''Try to inverse transform stored
        feature importances (either beta weights or
        automatically calculated feature importances)
        to their original space.

        .. warning ::

            | If there are any underlying non-recoverable
              transformations in the pipeline, this method
              will fail! For example, if a PCA was applied,
              then a reverse transformation cannot be computed.

        This method can be especially helpful when using :class:`Loader`.

        Returns
        --------
        inverse_fis : list of pandas Series
            | The inverse feature importances will be returned
              as a list, where each index of the list refers to
              a fold of the cross-validation, and each element
              of the list is either a pandas Series or a list
              of pandas Series (in the case of a categorical
              problem type where separate feature importances
              were calculated for each class).

            | If a :class:`Loader` was used, the returned Series
              may contain multi-dimensional arrays instead of scalar
              values, representing feature importances as transformed
              back into the original loaded space / shape.
        '''

        # As list of series or list of list of series
        fis = self._get_base_fis_list()

        inv_trans_fis = []
        for i, fi in enumerate(fis):

            # The estimator for this fold
            estimator = self.estimators[i]

            # Non-categorical case
            if isinstance(fi, pd.Series):
                inv_trans_fis.append(
                    estimator.inverse_transform_FIs(fi))

            # Categorical case
            else:
                cat_inv_fis =\
                    [estimator.inverse_transform_FIs(f) for f in fi]
                inv_trans_fis.append(cat_inv_fis)

        return inv_trans_fis

    def _get_val_fold_Xy(self, estimator, X_df, y_df, fold, just_model=True):

        # Get the X and y df's - assume self.val_indices stores
        # only subjects with non nan target variables
        X_val_df = X_df.loc[self.val_indices[fold]]
        y_val_df = y_df.loc[self.val_indices[fold]]

        # Base as array, and all feat names
        X_trans, feat_names = np.array(X_val_df), list(X_val_df)

        # Transform the X df, casts to array if just_model.
        if just_model:
            feat_names =\
                estimator.transform_feat_names(feat_names,
                                               encoders=self.encoders_)
            X_trans = estimator.transform(X_trans,
                                          transform_index=X_val_df.index)
            estimator = estimator._final_estimator

        return estimator, X_trans, np.array(y_val_df), feat_names

    def permutation_importance(self, dataset,
                               n_repeats=10, scorer='default',
                               just_model=True, return_as='dfs',
                               n_jobs=1, random_state='default'):
        '''This function computes the permutation feature importances
        from the base scikit-learn function
        :func:`sklearn.inspection.permutation_importance`

        Parameters
        -----------
        dataset : :class:`Dataset`
            The instance of the Dataset class originally passed to
            :func:`evaluate`. Note: if you pass a different dataset,
            you may get unexpected behavior.

        n_repeats : int, optional
            The number of times to randomly permute each feature.

            ::

                default = 10

        scorer : sklearn-style scoring, optional
            Scorer to use. It can be a single sklearn style str,
            or a callable.

            If left as 'default' will use the first scorer defined when
            evaluating the underlying estimator.

            ::

                default = 'default'

        just_model : bool, optional
            When set to true, the permutation feature importances
            will be computed using the final set of transformed features
            as passed when fitting the base model. This is reccomended
            behavior because it means that the features do not need to
            be re-transformed through the full pipeline to evaluate each
            feature. If set to False, will permute the features in the
            original feature space (which may be useful in some context).

            ::

                default = True

        return_as : {'dfs', 'raw'}, optional
            This parameter controls if calculated permutation
            feature importances should be returned as a DataFrame
            with column names as the corresponding feature names,
            or if it should be returned as a list with the raw
            output from each fold, e.g., sklearn Batch's with
            parameters 'importances_mean', 'importances_std'
            and 'importances'.

            If return as DataFrame is requested, then
            'importances_mean' and 'importances_std'
            will be returned, but not the raw 'importances'.

            ::

                default = 'dfs'


        n_jobs : int, optional
            The number of jobs to use for this function. Note
            that if the underlying estimator supports multiple jobs
            during inference (predicting), and the original
            problem_spec was set with multiple n_jobs then that original
            behavior will still hold, and you may wish to keep this
            parameter as 1. On the otherhand, if the base estimator
            does not use multiple jobs, passing a higher value here
            could greatly speed up computation.

            ::

                default = 1

        random_state : int, 'default' or None, optional
            Pseudo-random number generator to control the permutations
            of each feature.
            If left as 'default' then use the random state defined
            during the initial evaluation of the model. Otherwise, you may
            pass an int for a different fixed random state or None
            for explicitly no
            random state.

            ::

                default = 'default'
        '''

        # @TODO in case of just_model = False, won't pass along
        # transform_index correctly to scorer.

        from sklearn.inspection import permutation_importance

        self._estimators_check()

        # If default scorer, take the first one
        if scorer == 'default':
            first = list(self.ps.scorer)[0]
            scorer = self.ps.scorer[first]
            self._print('Using scorer:', first, level=1)

        # If default random_state use the one saved in
        # original problem spec.
        if random_state == 'default':
            random_state = self.ps.random_state

        # Get X and y from saved problem spec
        X, y = dataset.get_Xy(self.ps)

        # For each estimator
        all_fis, all_feat_names = [], []
        for fold, estimator in enumerate(self.estimators):

            # Get correct estimator, X_val, y_val and feat_names
            estimator, X_val, y_val, feat_names =\
                self._get_val_fold_Xy(estimator, X_df=X, y_df=y,
                                      fold=fold, just_model=just_model)
            all_feat_names.append(feat_names)

            # Run the sklearn feature importances.
            fis = permutation_importance(estimator, X_val, y_val,
                                         scoring=scorer, n_repeats=n_repeats,
                                         n_jobs=n_jobs,
                                         random_state=random_state)
            # Add to all fis
            all_fis.append(fis)

        # If raw, return as raw
        if return_as == 'raw':
            return all_fis

        # Otherwise, use return df
        mean_series, std_series = [], []
        for fis, feat_names in zip(all_fis, all_feat_names):
            mean_series.append(
                fi_to_series(fis['importances_mean'], feat_names))
            std_series.append(
                fi_to_series(fis['importances_std'], feat_names))

        # Return as sklearn bunch of DataFrames
        return Bunch(importances_mean=fis_to_df(mean_series),
                     importances_std=fis_to_df(std_series))

    def get_X_transform_df(self, dataset, fold=0, subjects='tr'):
        '''This method is used as a helper for getting the transformed
        input data for one of the saved models run during evaluate.

        Parameters
        -----------
        dataset : :class:`Dataset`
            The instance of the Dataset class originally passed to
            :func:`evaluate`. Note: if you pass a different dataset,
            you may get unexpected behavior.

        fold : int, optional
            The corresponding fold of the trained
            estimator to use.

        subjects : 'tr', 'val' or :ref:`Subjects`, optional
            The subjects data in which to return. As
            either special strings 'tr' for train subjects
            in the corresponding fold. Special str
            'val' for the validation subjects in the
            selected for or lastly any valid
            :ref:`Subjects` style input.

        Returns
        ----------
        X_trans_df : pandas DataFrame
            The transformed features in a DataFrame,
            according to the saved estimator from a fold,
            for the specified subjects.

            If kept as the default of subjects == 'tr',
            then these represent the feature values as
            passed to trained the actual model component
            of the pipeline.
        '''

        self._estimators_check()

        # Estimator from the fold
        estimator = self.estimators[fold]

        if subjects == 'tr':
            subjects = self.train_indices[fold]
        elif subjects == 'val':
            subjects = self.val_indices[fold]

        # Get feature names from fold
        feat_names = self.feat_names[fold]

        # Get as X
        X_fold, _ = dataset.get_Xy(problem_spec=self.ps,
                                   subjects=subjects)

        # Transform the data up to right before it gets passed to the
        # elastic net
        X_trans_fold = estimator.transform(X_fold)

        # Put the data in a dataframe with associated feature names
        return pd.DataFrame(X_trans_fold, columns=feat_names)

    def compare(self, other, rope_interval=[-0.01, 0.01]):
        '''This method is designed to perform a statistical comparison
        between the results from the evaluation stored in this object
        and another instance of :class:`BPtEvaluator`. The statistics
        produced here are explained in:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_grid_search_stats.html

        .. note::
            In the case that the sizes of the training and validation sets
            at each fold vary dramatically, it is unclear if this
            statistics are still valid.
            In that case, the mean train size and mean validation sizes
            are employed when computing statistics.

        Parameters
        ------------
        other : :class:`BPtEvaluator`
            Another instance of :class:`BPtEvaluator` in which
            to compare which. The cross-validation used
            should be the same in both instances, otherwise
            statistics will not be generated.

        rope_interval : list or dict of
            | This parameter allows for passing in a custom
                region of practical equivalence interval (or rope interval)
                a concept from bayesian statistics. If passed as
                a list, this should be a list with two elements, describing
                the difference in score which should be treated as two
                models or runs being practically equivalent.

            | Alternatively, in the case of multiple underlying
                scorers / metrics. A dictionary, where keys correspond
                to scorer / metric names can be passed with a separate
                rope_interval for each. For example:

            ::

                rope_interval = {'explained_variance': [-0.01, 0.01],
                                 'neg_mean_squared_error': [-1, 1]}

            This example would define separate rope regions depending
            on the metric.

            ::

                default = [-0.01, 0.01]

        Returns
        -------
        compare_df : pandas DataFrame
            | The returned DataFrame will generate separate rows
                for all overlapping metrics / scorers between the
                evaluators being compared. Further, columns with
                statistics of interest will be generated:

                - 'mean_diff'
                    The mean score minus other's mean score

                - 'std_diff'
                    The std minus other's std

            | Further, only in the case that the cross-validation
                folds are identical between the comparisons,
                the following additional columns will be generated:

                - 't_stat'
                    Corrected paired ttest statistic.

                - 'p_val'
                    The p value for the corrected paired ttest statistic.

                - 'better_prob'
                    The probability that this evaluated option is better than
                    the other evaluated option under a bayesian framework and
                    the passed value of rope_interval. See sklearn example
                    for more details.

                - 'worse_prob'
                    The probability that this evaluated option is worse than
                    the other evaluated option under a bayesian framework and
                    the passed value of rope_interval. See sklearn example
                    for more details.

                - 'rope_prob'
                    The probability that this evaluated option is equivalent to
                    the other evaluated option under a bayesian framework and
                    the passed value of rope_interval. See sklearn example
                    for more details.

        '''

        equal_cv = True

        # Make sure same number of folds
        if len(self.train_indices) != len(other.train_indices):
            equal_cv = False

        # Make sure subjects from folds line up
        for fold in range(len(self.train_indices)):
            if not np.array_equal(self.train_indices[fold],
                                  other.train_indices[fold]):
                equal_cv = False

            if not np.array_equal(self.val_indices[fold],
                                  other.val_indices[fold]):
                equal_cv = False

        # Only compute for the overlapping metrics
        overlap_metrics = set(list(self.mean_scores)).intersection(set(
            list(other.mean_scores)))

        for metric in overlap_metrics:
            if np.array_equal(self.scores[metric], other.scores[metric]):
                raise RuntimeError(
                    f'Cannot compare as scores are identical for {metric}.')

        # Init difference dataframe
        dif_df = pd.DataFrame(index=list(overlap_metrics))

        # Add base differences
        for metric in overlap_metrics:

            dif_df.loc[metric, 'mean_diff'] =\
                self.mean_scores[metric] - other.mean_scores[metric]

            dif_df.loc[metric, 'std_diff'] =\
                self.std_scores[metric] - other.std_scores[metric]

        # Only compute p-values if equal cv
        if equal_cv:
            for metric in overlap_metrics:

                # Grab scores and other info
                scores1 = np.array(self.scores[metric])
                scores2 = np.array(other.scores[metric])

                differences = scores1 - scores2
                n = len(scores1)
                df = n - 1

                # Use the mean train / test size
                n_train = np.mean([len(ti) for ti in self.train_indices])
                n_test = np.mean([len(ti) for ti in self.val_indices])

                # Frequentist Approach
                t_stat, p_val = compute_corrected_ttest(differences, df,
                                                        n_train, n_test)
                dif_df.loc[metric, 't_stat'] = t_stat
                dif_df.loc[metric, 'p_val'] = p_val

                # Bayesian
                t_post = t(df, loc=np.mean(differences),
                           scale=corrected_std(differences, n_train, n_test))

                # Passed as either list of two values or dict
                if isinstance(rope_interval, dict):
                    ri = rope_interval[metric]
                else:
                    ri = rope_interval

                worse_prob = t_post.cdf(ri[0])
                better_prob = 1 - t_post.cdf(ri[1])
                rope_prob =\
                    t_post.cdf(ri[1]) - t_post.cdf(ri[0])

                # Add to dif_df
                dif_df.loc[metric, 'better_prob'] = better_prob
                dif_df.loc[metric, 'worse_prob'] = worse_prob
                dif_df.loc[metric, 'rope_prob'] = rope_prob

        return dif_df



from lightgbm import LGBMRegressor, LGBMClassifier
from ..main.CV import BPtCV
from ..pipeline.helpers import proc_mapping
from sklearn.model_selection import train_test_split
import warnings


class BPtMixIn():

    _needs_mapping = True
    _needs_fit_index = True
    _needs_cat_inds = True

    def _get_categorical_feature(self, mapping):

        if not hasattr(self, 'cat_inds'):
            return 'auto'

        # If passed cat inds
        if self.cat_inds is not None:

            # Check for passed mapping, if found apply
            if mapping is not None:
                return proc_mapping(self.cat_inds, mapping)

            # Otherwise assume as is
            return self.cat_inds

        # Otherwise keep as default
        return 'auto'

    def _get_eval_set(self, X, y, fit_index=None):

        # Make sure there's early stop rounds set!
        if not hasattr(self, 'early_stopping_rounds'):
            return X, y, None
        if self.early_stopping_rounds is None:
            return X, y, None

        # Check for eval split
        if hasattr(self, 'eval_split'):

            # If any eval_split
            if self.eval_split is not None:
                if isinstance(self.eval_split, BPtCV):

                    # Get the cv_inds
                    train_inds, eval_inds =\
                        self.eval_split.get_split(
                            fit_index,
                            random_state=self.random_state)

                    # Index
                    X_train, y_train = X[train_inds], y[train_inds]
                    X_eval, y_eval = X[eval_inds], y[eval_inds]

                # Otherwise interpret as test size
                else:
                    X_train, X_eval, y_train, y_eval =\
                        train_test_split(X, y, test_size=self.eval_split,
                                         random_state=self.random_state)

                # Return
                return X_train, y_train, (X_eval, y_eval)

        return X, y, None

    def fit(self, X, y, mapping=None, fit_index=None, **kwargs):

        # Get rid of "other params"
        self._other_params = {}

        # Get categorical features based on if passed mapping
        categorical_feature = self._get_categorical_feature(mapping)

        # Proc eval set
        X_train, y_train, eval_set =\
            self._get_eval_set(X, y, fit_index=fit_index)

        # Check early stopping rounds:
        if hasattr(self, 'early_stopping_rounds'):
            early_stopping_rounds = self.early_stopping_rounds
        else:
            early_stopping_rounds = None

        # To avoid categorical feature warnings...
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # Call super fit
            super().fit(X_train, y_train,
                        eval_set=eval_set,
                        categorical_feature=categorical_feature,
                        early_stopping_rounds=early_stopping_rounds,
                        verbose=False,
                        **kwargs)


class BPtLGBMRegressor(BPtMixIn, LGBMRegressor):
    pass


class BPtLGBMClassifier(BPtMixIn, LGBMClassifier):
    pass



from .test_evaluate import get_fake_dataset
from ..input import (Model, Pipeline, Scaler, CV)
from ..funcs import evaluate
from ..BPtEvaluator import is_notebook
import pytest
import numpy as np


def test_is_notebook():

    assert not is_notebook()


def test_bpt_evaluator_compare_fail():

    pipe1 = Pipeline([Scaler('standard'), Model('linear')])
    pipe2 = Pipeline([Scaler('standard'), Model('dt')])

    dataset = get_fake_dataset()

    results1 = evaluate(pipeline=pipe1,
                        dataset=dataset,
                        progress_bar=False,
                        random_state=2,
                        cv=2)
    results2 = evaluate(pipeline=pipe2,
                        dataset=dataset,
                        progress_bar=False,
                        random_state=2,
                        cv=2)

    with pytest.raises(RuntimeError):
        results1.compare(results2)


def test_bpt_evaluator_compare():

    pipe1 = Pipeline([Scaler('standard'), Model('linear')])
    pipe2 = Pipeline([Model('dt')])

    dataset = get_fake_dataset()
    dataset['3'] = np.random.random(len(dataset))

    results1 = evaluate(pipeline=pipe1,
                        dataset=dataset,
                        progress_bar=False,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        random_state=2,
                        cv=2)

    dataset['3'] = np.random.random(len(dataset))
    results2 = evaluate(pipeline=pipe2,
                        dataset=dataset,
                        progress_bar=False,
                        random_state=2,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        cv=2)

    # Just in case, make sure dif results
    results2.mean_scores['explained_variance'] = .9
    results2.scores['explained_variance'] = [.8, 1]

    compare_df = results1.compare(results2)
    assert compare_df.shape == (2, 7)


def test_bpt_evaluator_compare_non_overlap_metric():

    pipe1 = Pipeline([Scaler('standard'), Model('linear')])
    pipe2 = Pipeline([Model('dt')])

    dataset = get_fake_dataset()
    dataset['3'] = np.random.random(len(dataset))

    results1 = evaluate(pipeline=pipe1,
                        dataset=dataset,
                        progress_bar=False,
                        scorer=['neg_mean_squared_error'],
                        random_state=2,
                        cv=2)

    # Just in case, make sure dif results
    dataset['3'] = np.random.random(len(dataset))
    results2 = evaluate(pipeline=pipe2,
                        dataset=dataset,
                        progress_bar=False,
                        random_state=2,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        cv=2)

    compare_df = results1.compare(results2)
    assert compare_df.shape == (1, 7)


def test_bpt_evaluator_compare_non_overlap_cv1():

    pipe1 = Pipeline([Scaler('standard'), Model('linear')])
    pipe2 = Pipeline([Model('dt')])

    dataset = get_fake_dataset()
    dataset['3'] = np.random.random(len(dataset))

    results1 = evaluate(pipeline=pipe1,
                        dataset=dataset,
                        progress_bar=False,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        random_state=2,
                        cv=2)

    # Just in case, make sure dif results
    dataset['3'] = np.random.random(len(dataset))
    results2 = evaluate(pipeline=pipe2,
                        dataset=dataset,
                        progress_bar=False,
                        random_state=2,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        cv=3)
    results2.mean_scores['explained_variance'] = .9
    results2.scores['explained_variance'] = [.8, 1, .9]

    compare_df = results1.compare(results2)
    assert compare_df.shape == (2, 2)


def test_bpt_evaluator_compare_non_overlap_cv2():

    pipe1 = Pipeline([Scaler('standard'), Model('linear')])
    pipe2 = Pipeline([Model('dt')])

    dataset = get_fake_dataset()
    dataset['3'] = np.random.random(len(dataset))

    results1 = evaluate(pipeline=pipe1,
                        dataset=dataset,
                        progress_bar=False,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        random_state=2,
                        cv=2)

    # Just in case, make sure dif results
    dataset['3'] = np.random.random(len(dataset))
    results2 = evaluate(pipeline=pipe2,
                        dataset=dataset,
                        progress_bar=False,
                        random_state=10,
                        scorer=['neg_mean_squared_error',
                                'explained_variance'],
                        cv=2)

    results2.mean_scores['explained_variance'] = .9
    results2.scores['explained_variance'] = [.8, 1]

    compare_df = results1.compare(results2)
    assert compare_df.shape == (2, 2)


def test_multiclass_get_preds_df():

    df = get_fake_dataset()
    df['3'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2]
    df.ordinalize('3', inplace=True)

    pipe = Pipeline([Model('linear')])

    results = evaluate(pipeline=pipe,
                       dataset=df,
                       progress_bar=False,
                       scorer='roc_auc_ovr',
                       cv=CV(splits=2))

    assert len(results.preds['predict']) == 2
    assert len(results.preds['predict_proba']) == 2

    assert len(results.preds['predict'][0]) == 10
    assert len(results.preds['predict_proba'][0]) == 10

    assert len(results.preds['predict_proba'][0][0]) == 3

    # Test get preds df
    r_df = results.get_preds_dfs()
    assert r_df[0].shape == (10, 8)
    assert r_df[0].shape == (10, 8)


def test_permutation_feature_importance():

    pipe = Pipeline([Scaler('standard'), Model('linear')])
    dataset = get_fake_dataset()
    results = evaluate(pipeline=pipe,
                       dataset=dataset,
                       progress_bar=False,
                       scorer='neg_mean_squared_error',
                       random_state=2,
                       cv=2)

    fis = results.permutation_importance(dataset, n_repeats=10)
    assert fis['importances_mean'].shape == (2, 2)
    assert fis['importances_std'].shape == (2, 2)



from scipy.stats import t
import numpy as np


def corrected_std(differences, n_train, n_test):
    """Corrects standard deviation using Nadeau and Bengio's approach.

    Parameters
    ----------
    differences : ndarray of shape (n_samples, 1)
        Vector containing the differences in the score metrics of two models.
    n_train : int
        Number of samples in the training set.
    n_test : int
        Number of samples in the testing set.

    Returns
    -------
    corrected_std : int
        Variance-corrected standard deviation of the set of differences.
    """
    n = n_train + n_test
    corrected_var = (
        np.var(differences, ddof=1) * ((1 / n) + (n_test / n_train))
    )
    corrected_std = np.sqrt(corrected_var)
    return corrected_std


def compute_corrected_ttest(differences, df, n_train, n_test):
    """Computes right-tailed paired t-test with corrected variance.

    Parameters
    ----------
    differences : array-like of shape (n_samples, 1)
        Vector containing the differences in the score metrics of two models.
    df : int
        Degrees of freedom.
    n_train : int
        Number of samples in the training set.
    n_test : int
        Number of samples in the testing set.

    Returns
    -------
    t_stat : float
        Variance-corrected t-statistic.
    p_val : float
        Variance-corrected p-value.
    """
    mean = np.mean(differences)
    std = corrected_std(differences, n_train, n_test)
    t_stat = mean / std
    p_val = t.sf(np.abs(t_stat), df)  # right-tailed t-test
    return t_stat, p_val



from .compare import Compare, Option
from copy import deepcopy, copy
from sklearn.base import BaseEstimator
from ..default.helpers import proc_input, coarse_any_obj_check
from ..util import conv_to_list, BPtInputMixIn

from ..main.input_operations import (Pipe, Select,
                                     ValueSubset, Duplicate)
from ..default.options.scorers import process_scorer
from .CV import get_bpt_cv
from ..pipeline.constructors import (LoaderConstructor, ImputerConstructor,
                                     ScalerConstructor, TransformerConstructor,
                                     FeatSelectorConstructor, ModelConstructor,
                                     CustomConstructor)

import warnings
from pandas.util._decorators import doc
from sklearn.utils.estimator_checks import check_estimator
from sklearn.pipeline import _name_estimators


def proc_all(base_obj):

    if base_obj is None:
        return

    proc_name(base_obj, 'obj')
    proc_name(base_obj, 'base_model')
    proc_name(base_obj, 'models')
    proc_name(base_obj, 'scorer')
    proc_name(base_obj, 'target_scaler')


def proc_name(base_obj, name):

    if hasattr(base_obj, name):
        obj = getattr(base_obj, name)

        if isinstance(obj, (list, tuple)):
            for o in obj:
                proc_all(o)

        elif hasattr(obj, 'obj'):
            proc_all(obj)

        elif isinstance(obj, str):
            setattr(base_obj, name, proc_input(obj))


def uniquify(obj):

    for p_str in ['obj', 'base_model', 'models', 'target_scaler']:

        if hasattr(obj, p_str):
            p = getattr(obj, p_str)

            # Recursive check first
            if isinstance(p, list):
                [uniquify(pp) for pp in p]
            elif hasattr(p, '_uniquify'):
                p._uniquify()

            # Then set with copy
            setattr(obj, p_str, copy(p))


class Params(BaseEstimator):

    def copy(self):
        '''This method returns a deepcopy of the base object.'''
        return deepcopy(self)

    def get_params(self, deep=True):

        params = super().get_params(deep=deep)

        if 'extra_params' in params and deep:
            extra_params = params.pop('extra_params')
            params.update(extra_params)

        return params

    def set_params(self, **params):

        # If no extra params
        if not hasattr(self, 'extra_params'):
            return super().set_params(**params)

        # Possible extra params case
        # Get base, no extra param keys
        base_keys = super().get_params().keys()

        # Set any base params
        base_params = {key: params[key] for key in params
                       if key in base_keys}
        super().set_params(**base_params)

        # Calculate any extra params to set
        extra_params = {key: params[key] for key in params
                        if key not in base_keys}

        # Update extra params
        self.extra_params.update(extra_params)

        return self


class Check():

    def _check_args(self):
        '''This method is used to ensure arguments are valid.
        It called automatically at the end of init.'''

        if hasattr(self, 'obj'):
            self._check_obj()

        if hasattr(self, 'params'):
            self._check_params()

        if hasattr(self, 'obj') and hasattr(self, 'params'):
            self._check_obj_and_params()

        if hasattr(self, 'base_model'):
            self._check_base_model()

        try:
            self._check_extra_args()
        except AttributeError:
            pass

    def _check_obj(self):

        obj = getattr(self, 'obj')

        if obj is None:
            raise RuntimeError('Passed obj cannot be None')

        if isinstance(obj, list) and not isinstance(obj, Pipe):
            raise RuntimeError('You may not pass a list of objs')

        # If str, then run a coarse general check
        # that won't handle if the object is of the wrong
        # problem type or for the wrong piece
        # but should catch typos
        if isinstance(obj, str):
            coarse_any_obj_check(obj)

    def _check_params(self):

        params = getattr(self, 'params')

        if params is None:
            setattr(self, 'params', 0)

    def _check_obj_and_params(self):

        obj = getattr(self, 'obj')
        params = getattr(self, 'params')

        if isinstance(obj, Pipe):

            if not isinstance(params, list):

                if params == 0:
                    new_params = [0 for i in range(len(obj))]
                    setattr(self, 'params', new_params)
                else:
                    raise IOError('obj is passed as Pipe, but params was not.',
                                  'Make sure',
                                  'params is either a list or Pipe ',
                                  'with the same length as',
                                  'self.obj')

            elif len(params) != len(obj):

                if params == [0]:
                    new_params = [0 for i in range(len(obj))]
                    setattr(self, 'params', new_params)
                else:
                    raise IOError('obj is passed as Pipe, Make sure',
                                  'params is either a list or Pipe with the ',
                                  'same length as',
                                  'self.obj')

    def _check_base_model(self):

        base_model = getattr(self, 'base_model')

        # None is okay
        if base_model is None:
            return

        # Assume select is okay too
        if isinstance(base_model, Select):
            return

        if not isinstance(base_model, Model):
            raise IOError('base_model must be either None or a valid '
                          'Model / Ensemble ',
                          'set of wrapper params!')


class Piece(Params, Check):
    '''This is the base piece in which :ref`pipeline_objects` inherit from. This
    class should not be used directly.'''

    # @TODO make this function work with custom objects
    def build(self, dataset, problem_spec='default', **problem_spec_params):
        '''This method is used to convert a single pipeline piece into
        the base sklearn style object used in the pipeline. This method
        is mostly used to investigate pieces and is not necessarily
        designed to produce independently usable pieces.

        For now this method will not work when the base
        obj is a custom object.

        Parameters
        -----------
        dataset : :class:`Dataset`
            The Dataset in which the pipeline should be initialized
            according to. For example, pipeline's can include Scopes,
            these need a reference Dataset.

            Something something

        problem_spec : :class:`ProblemSpec` or 'default', optional
            This parameter accepts an instance of the
            params class :class:`ProblemSpec`.
            The ProblemSpec is essentially a wrapper
            around commonly used
            parameters needs to define the context
            the model pipeline should be evaluated in.
            It includes parameters like problem_type, scorer, n_jobs,
            random_state, etc...

            See :class:`ProblemSpec` for more information
            and for how to create an instance of this object.

            If left as 'default', then will initialize a
            ProblemSpec with default params.

            ::

                default = "default"

        problem_spec_params : :class:`ProblemSpec` params, optional
            You may also pass any valid problem spec argument-value pairs here,
            in order to override a value in the passed :class:`ProblemSpec`.
            Overriding params should be passed in kwargs style, for example:

            ::

                func(..., problem_type='binary')

        Returns
        -------
        estimator : sklearn compatible estimator
            Returns the BPt-style sklearn compatible estimator
            version of this piece as converted to internally
            when building the pipeline

        params : dict
            Returns a dictionary with any parameter distributions
            associated with this object, for example
            this can be used to check what exactly
            pre-existing parameter distributions point
            to.

        Examples
        --------

        Given a dataset and pipeline piece (this can be any
        of the valid :ref:`api.pipeline_pieces` not just :class:`Model`
        as used here).

        .. ipython:: python

            import BPt as bp

            dataset = bp.Dataset()
            dataset['col1'] = [1, 2, 3]
            dataset['col2'] = [3, 4, 5]
            dataset.set_role('col2', 'target', inplace=True)
            dataset

            piece = bp.Model('ridge', params=1)
            piece

        We can call build from the piece

        .. ipython:: python

            estimator, params = piece.build(dataset)
            estimator
            params

        '''

        # @TODO also passing no dataset here
        # as long as all scopes are all, i.e., make
        # a fake dataset class and it should have a function
        # _get_data_inds, and _get_cols that throw errors
        # when invalid. Add this option to get_estimator also.

        from ..main.funcs import problem_spec_check

        # Get proc'ed problem spec
        ps = problem_spec_check(problem_spec, dataset=dataset,
                                **problem_spec_params)

        # Make sure dataset up to date
        dataset._check_sr()

        # Get constructor
        constructor =\
            self._constructor(spec=ps._get_spec(),
                              dataset=dataset, user_passed_objs={})

        # Return the objs and params
        objs, params = constructor.process(self)
        return objs[0], params

    def _get_param_names(self):

        param_names = super()._get_param_names()

        if hasattr(self, 'extra_params'):
            ep = getattr(self, 'extra_params')
            if isinstance(ep, dict) and len(ep) > 0:
                param_names += ['extra_params']

        return param_names

    def _uniquify(self):
        uniquify(self)


_piece_docs = {}

_piece_docs[
    "params"
] = """params : int, str or dict of :ref:`params<Params>`, optional
        | The parameter `params` can be used to set an associated distribution
            of hyper-parameters, fixed parameters or combination of.

        | Preset parameter options can be found distributions are
            listed for each choice of params with the corresponding
            obj at :ref:`Pipeline Options<pipeline_options>`.

        | More information on how this parameter
            works can be found at :ref:`Params`.

        ::

            default = 0
"""

_piece_docs[
    "scope"
] = """scope : :ref:`Scope`, optional
        | The `scope` parameter determines the subset of
            features / columns in which this object
            should operate on within the created pipeline. For example,
            by specifying scope = 'float', then
            this object will only operate on columns with scope
            float.

        | See :ref:`Scope` for more information on
            how scopes can be specified.
"""

_piece_docs[
    "extra_params"
] = """extra_params : :ref:`extra_params`
        | You may pass additional kwargs style arguments
            for this piece as :ref:`extra_params`. Any values
            passed here will be used to try and set that value
            in the requested obj.

        | Any parameter value pairs
            specified here will take priority over any
            set via params. For example, lets say in
            the object we are initializing, 'fake obj' it has a parameter
            called size, and we want it fixed as 10, we can specify that with:

        ::

            (obj='fake obj', ..., size=10)

        See :ref:`extra_params` for more information.
"""

_base_docs = _piece_docs.copy()

_piece_docs[
    "cache_loc"
] = """cache_loc : str, Path or None, optional
        | This parameter can optionally be set to a
            str or path representing the location in which
            this object will be cached after fitting.
            To skip this option, keep as the default argument of None.

        | If set, the python library joblib is used
            to cache a copy after fitting and in the case
            that a cached copy already exists will load from
            that copy instead of re-fitting the base object.

        ::

            default = None

    """

_bp_docs = _piece_docs.copy()

_piece_docs[
    "param_search"
] = """param_search : :class:`ParamSearch`, None, optional
        This parameter optionally specifies that this object
        should be nested with a hyper-parameter search.

        If passed an instance of :class:`ParamSearch`, the
        underlying object, or components of the underlying
        object (if a pipeline) must have atleast one valid
        hyper-parameter distribution to search over.

        By default this parameter is set to None, which
        specifies no parameter search.

        ::

            default = None

"""

_piece_docs[
    "target_scaler"
] = """target_scaler : :class:`Scaler`, None, optional
        | Can optionally pass an instance of
            :class:`Scaler` here to have properly
            nested target scaling / reverse scaling (before scoring)
            applied.

        .. warning::

            This parameter is still experimental.
            It has not been fully tested in
            complicated nesting cases, e.g., if Model is
            wrapping a nested Pipeline, this param will
            likely break.

        ::

            default = None

"""


@doc(**_base_docs)
class Loader(Piece):
    ''' Loader refers to transformations which operate on :ref:`data_files`.
    They in essence take in saved file locations and after some series
    of specified transformations pass on compatible features.

    The Loader object can operate in two ways. Either
    the Loader can define operations which are computed on
    single files independently, or load and pass on data
    to the defined `obj` as a list, where each element of
    the list is a subject's data. See parameter behav.

    Parameters
    ----------
    obj : str, custom obj or :class:`Pipe`
        | `obj` selects the base loader object to use, this can be either
          a str corresponding to
          one of the preset loaders found at :ref:`Loaders`.
          Beyond pre-defined loaders, users can
          pass in custom objects as long as they have functions
          corresponding to the correct behavior.

        | `obj` can also be passed as a :class:`Pipe`.
          See :class:`Pipe`'s documentation to
          learn more on how this works, and why you might want to use it.
          See :ref:`Pipeline Objects<pipeline_objects>` to read more
          about pipeline objects in general.

        | For example, the 'identity' loader will load in saved data at
          the stored file location, lets say they are 2d numpy arrays,
          and will return a flattened version
          of the saved arrays, with each data point as a feature.
          A more practical example
          might constitute loading in say 3D neuroimaging data,
          and passing on features as extracted by ROI.

    behav : 'single' or 'all', optional
        | The Loader object can operate under two different
          behaviors, corresponding to operations which can
          be done for each subject's Data File independently ('single')
          and operations which must be done using information
          from all train subject's Data Files ('all').

        | 'single' is the default behavior, if requested
          then the Loader will load each subject's Data File
          seperately and apply the passed `obj` fit_transform.
          The benefit of this method in contrast to 'all' is
          that only one subject's full raw data needs to be
          loaded at once, whereas with all, you must have enough
          avaliable memory to load all of the current training or
          validation subject's raw data at one. Likewise 'single'
          allows for caching fit_transform operations for each
          individual subject (which can then be more flexibly re-used).

        | Behavior 'all' is designed to work with base objects
          that accept a list of length n_subjects to their fit_transform
          function, where each element of the list will be that subject's
          loaded Data File. This behavior requires loading all data
          into memory, but allows for using information from the rest
          of the group split. For example we would need to set Loader
          to 'all' if we wanted to use
          :class:`nilearn.connectome.ConnectivityMeasure`
          with parameter kind = "tangent" as this transformer requires
          information from the rest of the loaded subjects when training.
          On the otherhand, if we used kind = "correlation",
          then we could use either behavior
          'all' or 'single' since "correlation" can be computed for each
          subject individually.

        ::

            default = 'single'

    {params}

    {scope}
        ::

            default = 'data file'

    cache_loc : str, Path or None, optional
        | Optional location in which to if set, the Loader transform
          function will be cached for each subject. These cached
          transformations can then be loaded for each subject when
          they appear again in later folds.

        .. warning ::

            If behav = 'all', then this parameter is currently
            skipped!

        Set to None, to ignore

        ::

            default = None

    fix_n_wrapper_jobs : int or False, optional
        Typically this parameter is left as default, but
        in special cases you may want to set this. It controls
        the number of jobs fixed for the Loading Wrapper.
        This parameter can be used to set that value.

        ::

            default = False

    {extra_params}

    Notes
    --------
    If obj is passed as :class:`Pipe`, see :class:`Pipe`
    for an example on how different
    corresponding params can be passed
    to each piece individually.

    See Also
    ---------
    Dataset.add_data_files : For adding data files to :class:`Dataset`.
    Pipe : An input helper class used with this object.

    Examples
    -----------
    A basic example is shown below:

    .. ipython:: python

        import BPt as bp
        loader = bp.Loader(obj='identity')
        loader

    This specifies that the :class:`BPt.extensions.Identity` loader be used
    (which just loads and flattens files).

    '''

    _constructor = LoaderConstructor

    def __init__(self, obj, behav='single', params=0, scope='data file',
                 cache_loc=None,
                 fix_n_wrapper_jobs=False, **extra_params):

        self.obj = obj

        # Make sure valid behav param
        if behav not in ['single', 'all']:
            raise RuntimeError('behav must be either "single" or "all".')

        self.behav = behav
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.extra_params = extra_params
        self.fix_n_wrapper_jobs = fix_n_wrapper_jobs

        self._check_args()


@doc(**_bp_docs)
class Imputer(Piece):
    '''This input object is used to specify imputation steps
    for a :class:`Pipeline`.

    If there is any missing data (NaN's), then an imputation strategy
    is likely necessary (with some expections, i.e., a final model which
    can accept NaN values directly). This object allows for defining
    an imputation strategy. In general, you should need at most two
    Imputers, one for all `float` type data and one for all
    categorical data. If there is no missing data, this piece will be skipped.

    Parameters
    ----------
    obj : str
        `obj` selects the base imputation strategy to use.
        See :ref:`Imputers` for all avaliable options.
        Notably, if 'iterative' is passed,
        then a base model must also be passed!

        See :ref:`Pipeline Objects<pipeline_objects>` to read more about
        pipeline objects in general.

    {params}

    {scope}
        | The main options that make sense for imputer are
            one for `float` data and one for `category` datatypes.
            Though in some cases other choices may make sense.

        | Note: If using iterative imputation you may want to carefully
            consider the scope passed. For example, while it may be beneficial
            to impute categorical and float features seperately, i.e., with
            different base_model_type's
            (categorical for categorical and regression for float), you must
            also consider that in predicting the missing values under
            this setup, the categorical imputer would not have access to
            to the float features and vice versa.
            In this way, you may want to either just
            treat all features as float, or
            instead of imputing categorical features, load missing
            values as a separate category - and then set the scope
            here to be 'all', such that the iterative imputer has
            access to all features.

        ::

            default = 'all'

    {cache_loc}

    base_model : :class:`Model`, :class:`Ensemble` or None, optional
        If 'iterative' is passed to obj, then a base_model is required in
        order to perform iterative imputation! The base model can be
        any valid :class:`Model` or :class:`Ensemble`

        ::

            default = None

    base_model_type : 'default' or Problem Type, optional
        In setting a base imputer model, it may be desirable to
        have this model have a different 'problem type', then your
        over-arching problem. For example, if performing iterative
        imputation on categorical features only, you will likely
        want to use a categorical predictor - but for imputing on
        float-type features, you will want to use a 'regression' type
        base model.

        Choices are 'binary', 'regression', 'categorical' or 'default'.
        If 'default', then the following behavior will be applied:
        If all columns within the passed scope of this Imputer object
        have scope / data type 'category', then the problem
        type for the base model will be set to 'categorical'.
        In all other cases, the problem type will be set to 'regression'.

        ::

            default = 'default'


    {extra_params}

    '''

    _constructor = ImputerConstructor

    def __init__(self, obj, params=0, scope='all',
                 cache_loc=None, base_model=None, base_model_type='default',
                 **extra_params):

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.base_model = deepcopy(base_model)
        self.base_model_type = base_model_type
        self.extra_params = extra_params

        self._check_args()


@doc(**_bp_docs)
class Scaler(Piece):
    '''The Scaler piece refers to
    a piece in the :class:`Pipeline` or :class:`ModelPipeline`,
    which is responsible for performing any sort of scaling or
    transformation on the data
    which doesn't require the target variable, and doesn't
    change the number of data points or features.
    These are typically transformations like feature scaling.

    Parameters
    ----------
    obj : str or custom obj
        | `obj` if passed a str selects a scaler from the preset defined
            scalers, See :ref:`Scalers`
            for all avaliable options. If passing a custom object, it must
            be a sklearn compatible
            transformer, and further must not require the target variable,
            not change the number of data
            points or features.
        |
        | See :ref:`Pipeline Objects<pipeline_objects>` to read more about
            pipeline objects in general.

    {params}

    {scope}
        ::

            default = 'float'

    {cache_loc}

    {extra_params}

    '''

    _constructor = ScalerConstructor

    def __init__(self, obj, params=0, scope='float',
                 cache_loc=None, **extra_params):

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.extra_params = extra_params

        self._check_args()


@doc(**_bp_docs)
class Transformer(Piece):
    ''' The Transformer is base optional component of the
    :class:`Pipeline` or :class:`ModelPipeline` classes.
    Transformers define any type of transformation to the loaded
    data which may change the number
    of features in a non-simple way (i.e., conceptually distinct
    from :class:`FeatSelector`, where you
    know in advance the transformation is just selecting a subset
    of existing features). These are
    transformations like applying Principle Component Analysis,
    or on the fly One Hot Encoding.

    Parameters
    ----------
    obj : str or custom_obj
        `obj` if passed a str selects from the avaliable class
        defined options for
        transformer as found at :ref:`Transformers`.

        If a custom object is passed as `obj`, it must be a
        sklearn api compatible
        transformer (i.e., have fit, transform, get_params
        and set_params methods, and further be
        cloneable via sklearn's clone function).

        See :ref:`Pipeline Objects<pipeline_objects>` to read more about
        pipeline objects in general.

    {params}

    {scope}

        | It may in some cases be useful to consider the use of
            :class:`Duplicate` here.

        ::

            default = 'data'

    {cache_loc}

    {extra_params}
    '''

    _constructor = TransformerConstructor

    def __init__(self, obj, params=0, scope='data', cache_loc=None,
                 **extra_params):

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.extra_params = extra_params

        self._check_args()


@doc(**_bp_docs)
class FeatSelector(Piece):
    ''' The FeatSelector class is a base piece of
    :class:`ModelPipeline` or :class:`Pipeline`, which is designed
    to preform feature selection.

    Parameters
    ----------
    obj : str or custom_obj
        | The `obj` parameter selects which
            feature selection strategy to use.

        | See :ref:`Feat Selectors` for all avaliable preset options
            and :ref:`Pipeline Objects<pipeline_objects>` to read more
            about pipeline objects in general.

        | Notably, if 'rfe' (recursive feature elimination) is passed, then a
            base model must also be passed!

    {params}

    {scope}
        ::

            default = 'all'

    {cache_loc}

    base_model : :class:`Model`, :class:`Ensemble` or None, optional
        | If 'rfe' is passed to obj, then a base_model is required in
            order to perform recursive feature elimination.

        | The base model can be any valid :class:`Model` or :class:`Ensemble`.

        ::

            default = None

    {extra_params}

    See Also
    ---------
    Scaler : For pieces which don't change the number of features.
    Transformer : For pieces which change the number of features
                  in different ways.

    '''

    _constructor = FeatSelectorConstructor

    def __init__(self, obj, params=0, scope='all',
                 cache_loc=None, base_model=None, **extra_params):

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.base_model = deepcopy(base_model)
        self.base_model_type = None
        self.extra_params = extra_params
        self._check_args()


@doc(**_piece_docs)
class Model(Piece):
    '''The Model class represents a
    base component of the :class:`Pipeline`
    / :class:`ModelPipeline`. This component
    is namely the estimator responsible for making
    predictions or classifications.

    obj : str or custom_obj
        | The passed object should be either
            a preset str indicator found at :ref:`Models`,
            or from a custom passed user model (compatible w/ sklearn api).

        .. note::

            The passed parameter should be either a single
            str or a custom object, not a list-like of either.
            In the case that an ensemble of :class:`Model` is
            needed, see :class:`Ensemble`.

        | See :ref:`Pipeline Objects<pipeline_objects>` to
            read more about pipeline objects in general.

    {params}

    {scope}
        ::

            default = 'all'

    {param_search}

    {target_scaler}

    {cache_loc}

    {extra_params}

    See Also
    ---------
    Ensemble : For an ensemble of Models.

    '''

    _constructor = ModelConstructor

    def __init__(self, obj, params=0, scope='all',
                 param_search=None, target_scaler=None,
                 cache_loc=None, **extra_params):

        self.obj = obj
        self.params = params
        self.scope = scope
        self.cache_loc = cache_loc
        self.param_search = param_search
        self.target_scaler = target_scaler
        self.extra_params = extra_params

        self._check_args()


@doc(**_piece_docs)
class Ensemble(Model):
    '''The Ensemble object is a
    :class:`ModelPipeline` (or :class:`Pipeline`) piece,
    designed to be passed as an estimator, the same
    as :class:`Model`. This class is used to create
    a variety of different ensemble based estimators.

    Parameters
    ----------
    obj : str
        | Each str passed to ensemble refers to a type of ensemble to train,
            based on also the passed input to the `models` parameter,
            and also the additional parameters passed when
            initializing this object.

        | See :ref:`Ensemble Types` to see all
            avaliable options for ensembles.

        .. warning::

            Passing custom objects here, while technically possible,
            is not currently full supported.
            That said, there are just certain assumptions that
            the custom object must meet in order to work, specifically,
            they should have similar input params to other similar existing
            ensembles, e.g., in the case the `single_estimator` is False
            and `needs_split` is also False, then the passed object needs
            to be able to accept an input parameter `estimators`,
            which accepts a list of (str, estimator) tuples.
            Whereas if needs_split is still False,
            but single_estimator is True, then the passed object needs
            to support an init param of `base_estimator`,
            which accepts a single estimator.

    models : :class:`Model`, :class:`Ensemble` or list of
        | The `models` parameter is designed to accept any single model-like
            pipeline parameter object, i.e.,
            :class:`Model` or even another :class:`Ensemble`.
            The passed pieces here will be used along with the
            requested ensemble object to
            create the requested ensemble.

        | See :class:`Model` for how to create a valid base model(s)
            to pass as input here.

    {params}

    {scope}
        ::

            default = 'all'

    {param_search}

    {target_scaler}

    base_model : :class:`Model`, None, optional
        | In the case that an ensemble method which has
            the parameter `final_estimator` (not base model),
            for example in the case of stacking,
            then you may pass a Model type
            object here to be used as that final estimator.

        | Otherwise, by default this will be left as None,
            and if the requested ensemble has the final_estimator
            parameter, then it will pass None to the object
            (which is typically for setting the default).

        ::

            default = None

    cv : :class:`CV` or None, optional
        | Used for passing custom nested internal
            CV split behavior to
            ensembles which employ splits, e.g., stacking.

        | The passed input can be either an instance
            of :class:`CV` or can be any valid
            scikit-learn style cv, e.g., the integer 5.

        ::

            default = None

    single_estimator : bool, optional
        | The parameter `single_estimator` is used to let the
            Ensemble object know if the passed `models`
            should be a single estimator, or in other words
            if the base ensemble object is expecting the input
            as just one estimator.

        | This parameter is used for ensemble types
            that requires an init param `base_estimator`.
            In the case that multiple models
            are passed to `models`, but `single_estimator` is True,
            then the models will automatically
            be wrapped in a voting ensemble,
            thus creating one single estimator.

        ::

            default = False

    n_jobs_type : 'ensemble' or 'models', optional
        | Valid options are either 'ensemble' or 'models'.

        | This parameter controls how the total n_jobs are distributed, if
            'ensemble', then the n_jobs will be used all in the ensemble object
            and every instance within the sub-models set to n_jobs = 1.
            Alternatively, if passed 'models', then the ensemble
            object will not be multi-processed, i.e.,
            will be set to n_jobs = 1, and the n_jobs will
            be distributed to each base model.

        | For example, if you are training a stacking regressor
            with n_jobs = 16, and you have 16+ models, then 'ensemble'
            is likely a good choice here. If instead you have only
            3 base models, and one or more of those 3 could benefit from
            a higher n_jobs, then setting n_jobs_type to 'models' might
            give a speed-up.

        ::

            default = 'ensemble'

    {extra_params}

    '''

    _constructor = ModelConstructor

    # @TODO add cache_loc here? Do the wrapper's support it?

    def __init__(self, obj, models,
                 params=0, scope='all',
                 param_search=None,
                 target_scaler=None,
                 base_model=None,
                 cv=None,
                 single_estimator=False,
                 n_jobs_type='ensemble',
                 **extra_params):

        self.obj = obj

        # Force passed models if not a list, into a list
        if not isinstance(models, list):
            models = [models]
        if isinstance(models, Select):
            models = [models]

        self.models = models
        self.params = params
        self.scope = scope
        self.param_search = param_search
        self.target_scaler = target_scaler
        self.base_model = base_model
        self.cv = cv
        self.single_estimator = single_estimator
        self.n_jobs_type = n_jobs_type
        self.extra_params = extra_params

        self._check_args()

    def _check_extra_args(self):

        if isinstance(self.models, list):
            for model in self.models:
                if not isinstance(model, Model):
                    raise IOError(
                        'All models must be valid Model/Ensemble !')

        else:
            if not isinstance(self.models, Model):
                raise IOError(
                    'Passed model in models must be a valid Model/Ensemble.')


class Custom(Piece):
    '''Wrapper around pieces passed
    as valid sklearn estimator.'''

    _constructor = CustomConstructor

    def __init__(self, step):

        # Save step
        self.step = step

    def _get_step(self):

        # If already tuple return as is
        if isinstance(self.step, tuple):
            step = self.step

        # Otherwise, get as tuple
        else:
            step = _name_estimators([self.step])[0]

        # Change name if needed
        name, obj = step[0], step[1]

        # Return step
        return (name, obj)

    def __repr__(self):
        return self.step.__repr__()

    def __str__(self):
        return self.step.__str__()


class ParamSearch(Params):
    '''ParamSearch is special input object designed to be
    used with :class:`ModelPipeline` or :class:`Pipeline` that
    is used in order to define a hyperparameter search strategy.

    | When passed to :class:`Pipeline`, its search strategy is
      applied in the context of any set :ref:`Params` within the base pieces.
      Specifically, there must be at least one parameter search
      somewhere in the object ParamSearch is passed!
      All backend hyper-parameter searches make use of the
      <https://github.com/facebookresearch/nevergrad>`_ library.

    Parameters
    ----------
    search_type : str, optional
        | The type of nevergrad hyper-parameter search to conduct. See
            :ref:`Search Types<search_type_options>` for all avaliable options.

        | You may pass 'grid' here in addition to the
            supported nevergrad searches. This will use sklearn's
            GridSearch. Note in this case some of the other parameters
            are ignored, these are: weight_scorer, mp_context, dask_ip,
            memmap_X, search_only_params

        ::

            default = 'RandomSearch'

    cv : :class:`CV` or 'default', optional
        | The hyper-parameter search works by internally
            evaluating each combination of parameters. In
            order to internally evaluate a set of parameters,
            there must be some type of cross-validation defined.
            This parameter is used to represent the choice of
            cross-validation to use. The set of parameters which
            achieves the highest average score across the folds
            defined here will be selected.

        | Passed input should be an instance of :class:`CV`.
            If left as the custom str 'default', then a
            :class:`CV` object will be initialized and used
            with just the default parameters
            (which is a once repeated 3-fold cross-validation).

        ::

            default = 'default'

    n_iter : int, optional
        | This parameter represents the number of different
            hyper-parameters to try. It can also be thought of
            as the budget given to the underlying search algorithm.

        | How well a hyper-parameter search works
            (i.e., the quality of the chosen parameters) and how long
            it takes to run, are quite dependent on both this parameter
            and the passed cv strategy. In general, if too few
            choices are provided the algorithm will likely not select high
            performing hyperparameters, and alternatively
            if too high a value/budget is set, then you may find
            overfit/non-generalize hyper-parameter choices.

        | Other factors which might influence the 'right' number of
            `n_iter` to specify are:

            - `search_type`
                Depending on the underlying search type, it may
                take a bigger or smaller budget
                on average to find a good set of hyper-parameters

            - The dimension of the underlying search space
                If you are only optimizing a few, say 2,
                underlying parameter distributions,
                this will require a far smaller budget then say a really
                high dimensional search space.

            - The CV strategy
                The CV strategy defined via `cv`
                may make it easier or harder to overfit when searching
                for hyper-parameters, thus conceptually
                a good choice of cross-validation strategy can
                serve to increase the number `n_iter` you can use
                before overfitting or alternatively a bad choice may limit it.

            - Number of data points / subjects
                Along with CV strategy, the number of data points/subjects
                will greatly influence
                how quickly you overfit, and therefore a
                good choice of `n_iter`.

        ::

            default = 10

    scorer : str or 'default', optional
        | In order for a set of hyper-parameters to be evaluated,
            a single scorer must be defined. In the case
            that multiple scorer's are passed here, the first
            one will be used and the others ignored.

        | For a full list of supported scorers please view the
            scikit-learn docs at:
            https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

        .. note::

            If selecting a custom (i.e., anything but 'default'),
            be careful to make sure to select an appropriate scorer for
            the underlying problem type.

        | If left as 'default', a reasonable scorer based on the
            problem type is used.

        - 'regression'  : 'explained_variance'
        - 'binary'      : 'matthews'
        - 'categorical' : 'matthews'

        ::

            default = 'default'

    weight_scorer : bool or 'default', optional
        | The `weight_scorer` parameter allows for
            optionally weighting the scores by
            the number of subjects within each validation fold.
            The mean score is then if set to True, a weighted average
            instead.

        | This parameter is typically only useful in the case
            where the folds vary dramatically by size (e.g., in
            the case of a leave-out-group cross-validation, where
            the groups vary in size).

        ::

            default = False

    mp_context : str, optional
        | When a hyper-parameter search is launched there are different
            ways through python that the multi-processing can be launched
            (assuming n_jobs > 1). Depending on the system running the code,
            some options may be more reliable than others.

        | Valid options are:

            - 'loky': Create and use the python library loky backend.

            - 'fork': Python default fork mp_context

            - 'forkserver': Python default forkserver mp_context

            - 'spawn': Python default spawn mp_context

        | New as of version 1.3.6+, the 'loky' backend will be used,
            which is quite reliable across a number of systems and
            shouldn't really need to be changed.

        ::

            default = 'loky'

    n_jobs : int or 'default', optional
        | This parameter can be set in the case that
            a specific number of jobs (i.e., processors)
            should be used to run this parameter search.

        | If left as the default value of 'default'
            then the choice of n_jobs will be inherited
            through the context in which whatever
            object this search is associated with is
            used. This is typically a good choice,
            and this value can be left as 'default'.

        ::

            default = 'default'

    random_state : int, None or 'default', optional
        | If left as 'default', the random_state
            as set through an associated :class:`ProblemSpec`
            will be used. Otherwise, you may specify a
            specific value here. Either an integer representing
            a fixed random state or None to specify a
            new random state each time.

        ::

            default = 'default'


    dask_ip : str or None, optional
        If None, the default, then ignore this parameter.
        Otherwise, this parameter represents experimental Dask support.
        In this case the passed parameter should be a string representing
        the ip of a created dask cluster. A dask Client object will then
        be created and passed this ip in order to connect to the cluster.

        .. warning::

            This functionality is still experimental, and
            will not work if the underlying search_type is 'grid'.

        Built in to using dask to evaluate each combination of parameters
        is pre-scattering the data to each cluster node.

        ::

            default = None

    memmap_X : bool, optional
        | When passing large memory arrays in each parameter search,
            it can be useful as a memory reduction technique to pass
            numpy memmap'ed arrays. This solves an issue where
            the loky backend will not properly pass too large arrays.

        .. warning::

            This option can slow down code a large amount, and typically
            should not be used unless out of memory errors are encountered.
            This option will be skipped if the underlying
            search_type is 'grid', or is using `dask_ip` also.
            In the case of using `dask_ip` large data will be
            pre-scattered instead anyway.

        ::

            default = False

    search_only_params : dict or None, optional
        | In some rare cases, it may be the case that you want
            to specify that certain parameters be passed only during
            the nested parameter searches. A dict of parameters
            can be passed here to accomplish that. For example,
            if passing:

            ::

                search_only_params = {'svm classifier__probability': False}

        | Assuming that the default / selecting parameter for this svm
            classifier for probability is True by default, then only when
            exploring nested hyper-parameter options will probability be set
            to False, but when fitting the final model with the best
            parameters found from the search, it will revert back to
            the default (i.e., in this case probability = True).

        .. note::

            This may be difficult for non-advanced users to
            use, as you must pass parameters exactly as how
            they are represented internally. Using build
            on the piece of interest may be helpful in figuring
            out what this looks like for a specific piece and parameter.

        | To ignore this parameter / option.
            simply keep the default value of None.

        ::

            default = None

    verbose : int, optional
        Controls the verbosity of the search, where
        the higher value set, the more messages will be printed.
        By default no verbosity(i.e., verbose=0) will be used.

        ::

            default = 0

    progress_loc : None or str, optional
        This is an optional parameter. If set
        to non-None, then it should be passed a
        str representing the location of a text file
        to append to after every completed parameter.

        ::

            default = None
    '''

    def __init__(self, search_type='RandomSearch', cv='default',
                 n_iter=10, scorer='default', weight_scorer=False,
                 mp_context='loky', n_jobs='default',
                 random_state='default',
                 dask_ip=None, memmap_X=False,
                 search_only_params=None, verbose=0, progress_loc=None):
        self.search_type = search_type

        if cv == 'default':
            cv = CV()
        self.cv = cv

        # @TODO allow passing single arguments, e.g. splits
        # as param and have it set within cv

        self.n_iter = n_iter
        self.scorer = scorer
        self.weight_scorer = weight_scorer
        self.mp_context = mp_context
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.dask_ip = dask_ip
        self.memmap_X = memmap_X
        self.verbose = verbose
        self.progress_loc = progress_loc
        self.search_only_params = search_only_params

        self._check_args()

    def _as_dict(self, ps):

        params = self.get_params()

        if self.random_state == 'default':
            params['random_state'] = ps.random_state

        if self.n_jobs == 'default':
            params['n_jobs'] = ps.n_jobs

        params['scorer'] = process_scorer(self.scorer,
                                          ps.problem_type)

        if self.search_only_params is None:
            params['search_only_params'] = {}

        return params

    def _check_args(self):

        if isinstance(self.scorer, list):
            raise IOError('scorer within Param Search cannot be list-like')
        if isinstance(self.weight_scorer, list):
            raise IOError(
                'weight_scorer within Param Search cannot be list-like')


def check_if_sklearn_step(step):

    # Skip def. not valid cases
    if isinstance(step, (Piece, Pipeline, BPtInputMixIn, str)):
        return False

    # If Tuple
    if isinstance(step, tuple):
        est = step[1]

    # Otherwise assume directly as obj
    else:
        est = step

    # Use sklearn base check
    try:
        check_estimator(est)
    except AttributeError:
        return False

    return True


def add_nested_deep_params(params):
    '''When passed shallow params, get
    nested deep params w/ possibility for
    list.'''

    for param_key in list(params):
        value = params[param_key]

        # If parameter value is list
        if isinstance(value, list):
            for i, step in enumerate(value):

                # Compare case
                if isinstance(step, Compare):
                    params[param_key + '__' + str(i)] = step
                    continue

                # Skip non objects
                if not hasattr(step, 'get_params'):
                    continue

                # Get step params
                step_params = step.get_params(deep=True)

                # Add parameter
                for key in step_params:
                    new_key = param_key + '__' + str(i) + '__' + key
                    params[new_key] = step_params[key]

        # Otherwise, check for nested
        else:

            # Skip non objects
            if not hasattr(value, 'get_params'):
                continue

            # Get value params if object
            value_params = value.get_params(deep=True)

            for key in value_params:
                params[param_key + '__' + key] = value_params[key]


_pipeline_docs = {}
_pipeline_docs['param_search'] = _piece_docs['param_search']

_pipeline_docs['cache_loc'] = """cache_loc : Path str or None, optional
        Optional parameter specifying a directory
        in which full BPt pipeline's should
        be cached after fitting. This should be
        either left as None, or passed a str representing
        a directory in which cached fitted pipeline should be saved.

        ::

            default = None
"""

_pipeline_docs['verbose'] = """verbose : int, optional
        If greater than 0, print statements about
        the current progress of the pipeline during fitting.

        ::

            default = 0
"""


@doc(**_pipeline_docs)
class Pipeline(Params):
    '''This class is used to create flexible BPt style pipeline's.

    See :class:`ModelPipeline` for an alternate version of this class
    which enforces a strict ordering on how pipeline pieces can be set, and
    also includes a number of useful default behaviors.

    Parameters
    -----------
    steps : list of :ref:`Pipeline Objects<pipeline_objects>`
        | Input here is a list of :ref:`Pipeline Objects<pipeline_objects>`
            or custom valid sklearn-compliant objects / pipeline steps.
            These can be in any order as long as there is only
            one :class:`Model` or :class:`Ensemble`
            and it is at the end of the list, i.e., the last step.
            This constraint excludes any nested models.

        | See below for example usage.

    {param_search}

    {cache_loc}

    {verbose}

    Notes
    -------
    This class differs from :class:`sklearn.pipeline.Pipeline`
    most drastically in that this class is not itself directly a
    sklearn-complaint estimator (i.e., an object with fit and predict methods).
    Instead, this object represents a flexible set of input pieces,
    that can all vary depending on the eventual :class:`Dataset`
    and :class:`ProblemSpec` they are used in the context of.
    This means that instances of this class can be easily
    re-used across different data and setups, for example
    with different underlying problem_types (running a binary and then
    a regression version).

    Examples
    ----------
    The base behavior is to use all valid Pipeline objects,
    for example:

    .. ipython:: python

        pipe = bp.Pipeline(steps=[bp.Imputer('mean'),
                                  bp.Scaler('robust'),
                                  bp.Model('elastic')])
        pipe

    | This would creates a pipeline with mean imputation, robust scaling
      and an elastic net, all using the BPt style custom objects.

    | This object can also work with :class:`sklearn.pipeline.Pipeline`
      style steps. Or a mix of BPt style and sklearn style, for example:

    .. ipython:: python

        from sklearn.linear_model import Ridge

        pipe = bp.Pipeline(steps=[bp.Imputer('mean'),
                                  bp.Scaler('robust'),
                                  ('ridge regression', Ridge())])
        pipe

    You may also pass sklearn objects directly instead of as
    a tuple, i.e., in the :func:`sklearn.pipeline.make_pipeline`
    input style. For example:

    .. ipython:: python

        from sklearn.linear_model import Ridge
        pipe = bp.Pipeline(steps=[Ridge()])
        pipe

    .. note::

        Passing objects as sklearn-style
        ensures they have essentially a scope of 'all'
        and no associated hyper-parameter distributions.

    '''
    def __init__(self, steps, param_search=None, cache_loc=None, verbose=0):
        self.steps = steps
        self.param_search = param_search
        self.cache_loc = cache_loc
        self.verbose = verbose

        self._proc_checks()

    def _proc_checks(self):

        # Check for any custom passed steps
        self._check_for_custom()

        # Check last step
        self._check_last_step()

        # Validate rest of input
        self._validate_input()

        # Proc duplicates, then flatten
        self.steps = self._proc_duplicates(self.steps)
        self._flatten_steps()

        # Uniquify
        self._uniquify()

        # Proc input
        self.steps = self._proc_input(self.steps)

    def _check_for_custom(self):

        for i, step in enumerate(self.steps):

            # If is custom estimator, wrap in Custom
            if check_if_sklearn_step(step):
                self.steps[i] = Custom(step)

    def _check_last_step(self):

        if not isinstance(self.steps, list):
            raise TypeError('steps must be a list!')

        if isinstance(self.steps, Select):
            raise TypeError('steps cannot be Select')

        if len(self.steps) == 0:
            raise IndexError('steps cannot be empty!')

        # If last step is str, conv to Model
        if isinstance(self.steps[-1], str):
            self.steps[-1] = Model(self.steps[-1])

        # Make sure model
        self._check_is_model(self.steps[-1])

    def _check_is_model(self, step):

        # If bpt mixin, extra cases
        if isinstance(step, BPtInputMixIn):

            if isinstance(step, Compare):
                for option in step.options:
                    assert isinstance(option, Option)

            elif isinstance(step, Select):
                [self._check_is_model(o) for o in step]

            else:
                raise RuntimeError(repr(step) + ' is not a valid Model')

        # If last step isn't model - or custom, raise error
        if not isinstance(self.steps[-1], (Model, Pipeline,
                                           Custom, BPtInputMixIn)):
            raise RuntimeError(repr(step) + ' is not a valid Model')

    def _validate_input(self):
        '''Make sure all steps are Pieces, and param
        search is valid if not None'''

        # Validate steps
        for step in self.steps:

            if not isinstance(step, (Piece, Pipeline, Compare, Select)):
                raise RuntimeError('passed step:' + repr(step) +
                                   ' is not a valid Pipeline Piece / '
                                   'input wrapper')

            step._check_args()

        # Validate param search
        if self.param_search is not None:
            if not isinstance(self.param_search, ParamSearch):
                raise RuntimeError(repr(self.param_search) + 'is not a'
                                   ' ParamSearch.')

            # Check input args in case anything changed
            self.param_search._check_args()

    def _uniquify(self):

        # First call recursively
        for step in self.steps:
            if isinstance(step, (Piece, Pipeline, Compare, Select)):
                step._uniquify()

        # Then set steps with copy of steps
        self.steps = [copy(step) for step in self.steps]

    def _check_args(self):

        for step in self.steps:
            step._check_args()

    def _flatten_steps(self):

        # If just one element cast to list and end
        if not isinstance(self.steps, list):
            self.steps = [self.steps]
            return
        if isinstance(self.steps, Select):
            self.steps = [self.steps]
            return

        steps = []
        for step in self.steps:
            if isinstance(step, list) and not isinstance(step, Select):
                for sub_step in step:
                    steps.append(sub_step)
            else:
                steps.append(step)

        self.steps = steps

    def _proc_duplicates(self, params):

        if isinstance(params, list):
            new_params = deepcopy(params)
            del new_params[:]

            if len(params) == 1:
                return self._proc_duplicates(params[0])

            for p in params:
                new_params.append(self._proc_duplicates(p))

            return new_params

        elif hasattr(params, 'scope'):
            scopes = getattr(params, 'scope')
            new_params = []

            if isinstance(scopes, Duplicate):
                for scope in scopes:
                    new = deepcopy(params)
                    new.scope = scope
                    new_params.append(new)

                return new_params

        return params

    def _proc_input(self, params):

        for p in params:
            if isinstance(p, list):
                self._proc_input(p)
            else:
                proc_all(p)

        return params

    def _get_steps(self):
        return self.steps

    def _check_imputers(self, is_na):

        # If no NaN set to none
        if not is_na:
            self.imputers = None

    def get_params(self, deep=True):

        # Base behav get top level
        params = super().get_params(deep=False)

        # Then proc if deep
        if deep:
            add_nested_deep_params(params)

        return params

    def set_params(self, **params):

        for key in params:
            value = params[key]

            # Get last key
            split_key = key.split('__')
            last_key = split_key[-1]

            obj = self
            for key_piece in split_key[:-1]:

                # If int
                try:
                    indx = int(key_piece)
                    obj = obj[indx]

                # If parameter
                except ValueError:
                    obj = getattr(obj, key_piece)

            # Set value via set params if avaliable / not self
            if obj != self and hasattr(obj, 'set_params'):
                obj.set_params(**{last_key: value})
                return self

            # Check if last key is int
            try:
                indx = int(last_key)
                obj[indx] = value

            except ValueError:
                setattr(obj, last_key, value)

            return self


class ModelPipeline(Pipeline):
    '''The ModelPipeline class is used to create BPtPipeline's.
    The ModelPipeline differs from :class:`Pipeline`
    in that it enforces a simplification on the ordering of
    pieces, representing the typical order in which they might appear.
    See :class:`Pipeline` for a more flexible version of this class
    which does not enforce any ordering.

    The order enforced, which follows the order of the input arguments, is:

    1. loaders,
    2. imputers
    3. scalers
    4. transformers
    5. feat_selectors
    6. model

    For each parameter, which the exception of model,
    you may pass either one instance of that piece
    or a list of that piece. In the case that
    a list is passed, then it will be treated as
    a sequential set of steps / transformation
    where the output from each element of the
    list if passed on to the next as input.

    Parameters
    ----------
    loaders : :class:`Loader` or list of, optional
        Each :class:`Loader` refers to transformations
        which operate on loaded Data_Files
        See :class:`Loader`.

        You may wish to consider using the
        :class:`Pipe` input class
        when creating a single :class:`Loader` obj

        Passed loaders can also be wrapped in a
        :class:`Select` wrapper, e.g., as either

        ::

            # Just passing select
            loaders = Select([Loader(...), Loader(...)])

            # Or nested
            loaders = [Loader(...), Select([Loader(...), Loader(...)])]

        In this way, most of the pipeline objects can accept lists,
        or nested lists with param wrapped, not just loaders!

        ::

            default = None

    imputers : :class:`Imputer`, list of or None, optional
        If there is any missing data (NaN's) that have been kept
        within the input data, then an imputation strategy should likely
        be defined. This param controls what kind of
        imputation strategy to use.

        See :class:`Imputer`.

        You may also pass a value of 'default' here, which in
        the case that any NaN data is present within the training
        set, the set of two imputers:

        ::

            'default' == [Imputer('mean', scope='float'),
                          Imputer('median', scope='category')]

        Will be used. Otherwise, if there is no NaN present
        in the input data, no Imputer will be used.

        ::

            default = 'default'

    scalers : :class:`Scaler`, list of or None, optional
        Each :class:`Scaler` refers to any potential data scaling where a
        transformation on the data
        (without access to the target variable) is
        computed, and the number of features or
        data points does not change.

        See :class:`Scaler`.

        By keeping the default behavior, with parameter, 'default',
        standard scaling is applied,
        which sets each feature to have a std of 1 and mean 0.

        ::

            default = 'default'

    transformers : :class:`Transformer`, list of or None, optional
        Each :class:`Transformer` defines a type of transformation to
        the data that changes the number of
        features in perhaps non-deterministic
        or not simply removal
        (i.e., different from feat_selectors), for example
        applying a PCA, where both the number of features change, but
        also the new features do not 1:1 correspond to the original
        features. See :class:`Transformer` for more information.

        ::

            default = None

    feat_selectors : :class:`FeatSelector`, list of or None, optional
        Each :class:`FeatSelector` refers to an optional
        feature selection stage of the Pipeline.

        See :class:`FeatSelector`.

        ::

            default = None

    model : :class:`Model`, :class:`Ensemble`, optional
        This parameter accepts one input of type
        :class:`Model` or :class:`Ensemble`. Though,
        while it cannot accept a list (i.e., no sequential behavior), you
        may still pass Input Type wrapper like :class:`Select` to perform
        model selection via param search.

        See :class:`Model` for more information on how to specify a single
        model to BPt, and :class:`Ensemble` for information on how
        to build an ensemble of models.

        This parameter cannot be None. In the default
        case of passing 'default', a ridge regression
        is used.

        ::

            default =  'default'

    param_search : :class:`ParamSearch` or None, optional
        A :class:`ParamSearch` can be provided specifying that
        a nested hyper-parameter search be constructed across
        all valid param distributions (as set within each piece).

        If a parameter search is passed, it is required
        that atleast one piece have a relevant hyper-parameter
        distribution (i.e., as set in a pieces `params`).

        If param search is None, any defined parameter distributions
        will be ignored, though any static parameters specified via
        `params` will still be set.

        Note: The input wrapper :class:`Select`,
        is acts a hyper-parameter distribution
        and therefore requires a search to be passed here.

        ::

            default = None

    cache_loc : Path str or None, optional
        Optional parameter specifying a directory
        in which full BPt pipeline's should
        be cached after fitting. This may be useful
        in some contexts. If desired,
        pass the location of the directory in
        which to store this cache.

        ::

            default = None

    verbose : int, optional
        If greater than 0, print statements about
        the current progress of the pipeline during fitting.

        ::

            default = 0

        '''

    @property
    def fixed_piece_order(self):
        '''ModelPipeline has a fixed order in which
        pieces are constructed, this is:

            1. loaders,
            2. imputers
            3. scalers
            4. transformers
            5. feat_selectors
            6. model

        '''
        return ['loaders', 'imputers', 'scalers',
                'transformers', 'feat_selectors', 'model']

    def __init__(self,
                 loaders=None, imputers='default',
                 scalers='default', transformers=None,
                 feat_selectors=None,
                 model='default',
                 param_search=None,
                 cache_loc=None,
                 verbose=0):

        if isinstance(loaders, str):
            loaders = Loader(loaders)
        self.loaders = loaders

        if imputers == 'default':
            imputers = Imputer('default')
        elif isinstance(imputers, str):
            imputers = Imputer(imputers)
        self.imputers = imputers

        if isinstance(scalers, str):
            if scalers == 'default':
                scalers = Scaler('standard')
                if verbose > 0:
                    print('Passed default scalers, setting to:',
                          scalers)
            else:
                scalers = Scaler(scalers)

        self.scalers = scalers

        if isinstance(transformers, str):
            transformers = Transformer(transformers)
        self.transformers = transformers

        if isinstance(feat_selectors, str):
            feat_selectors = FeatSelector(feat_selectors)
        self.feat_selectors = feat_selectors

        if model == 'default':
            model = Model('ridge')
            if verbose > 0:
                print('Passed default model, setting to:', model)

        elif isinstance(model, str):
            model = Model(model)
        self.model = model

        self.param_search = param_search
        self.cache_loc = cache_loc
        self.verbose = verbose

        # Perform all preproc on input which can be run
        # more then once, these are essentially checks on the input
        self._proc_checks()

    def _proc_all_pieces(self, func):
        '''Helper to proc all pieces with a function
        that accepts a list of pieces'''

        for param_name in self.fixed_piece_order:
            params = getattr(self, param_name)

            if params is None:
                new_params = params

            elif isinstance(params, list):
                new_params = func(params)

            # If not list-
            else:
                new_params = func([params])
                if isinstance(new_params, list) and len(new_params) == 1:
                    new_params = new_params[0]

            setattr(self, param_name, new_params)

    def _uniquify(self):

        for param_name in self.fixed_piece_order:
            params = getattr(self, param_name)

            # Call recursively
            if isinstance(params, list):
                [p._uniquify() for p in params]
            elif hasattr(params, '_uniquify'):
                params._uniquify()

            # Then set params with copy
            setattr(self, param_name, copy(params))

    def _check_args(self, params):

        for p in params:
            if isinstance(p, list):
                self._check_args(p)
            else:
                p._check_args()

        return params

    def _check_imputers(self, is_na):

        # If no NaN set to none
        if not is_na:
            self.imputers = None

        elif isinstance(self.imputers, Imputer):
            if self.imputers.obj == 'default':
                self.imputers = [Imputer('mean', scope='float'),
                                 Imputer('median', scope='category')]

                if self.verbose > 0:
                    print('Default passed, setting imputers to:',
                          self.imputers)

        return

    def _proc_checks(self):

        try:
            self.model = self.models
            print('Passed models, set as model')
        except AttributeError:
            pass

        try:
            self.transformers = self.transformer
            print('Passed transformer, set as transformers')
        except AttributeError:
            pass

        try:
            self.imputers = self.imputer
            print('Passed imputer, set as imputers')
        except AttributeError:
            pass

        try:
            self.feat_selectors = self.feat_selector
            print('Passed feat_selector, set as feat_selectors')
        except AttributeError:
            pass

        try:
            self.scalers = self.scaler
            print('Passed scaler, set as scalers')
        except AttributeError:
            pass

        try:
            self.loaders = self.loader
            print('Passed loader, set as loaders')
        except AttributeError:
            pass

        to_check = ['ensemble', 'scorer', 'metric',
                    'feature_selector', 'feature_selectors']
        for p in to_check:
            if hasattr(self, p):
                print('Warning: ModelPipeline user set param', p,
                      ' was set,',
                      'but will have no effect as it is not',
                      ' a valid parameter!')

        # Uniquify
        self._uniquify()

        # Check for duplicate scopes
        self._proc_all_pieces(self._proc_duplicates)

        # Proc input
        self._proc_all_pieces(self._proc_input)
        proc_all(self.param_search)

        # Double check input args in case something changed
        self._proc_all_pieces(self._check_args)

        # Proc param search if not None
        if self.param_search is not None:
            self.param_search._check_args()

    def _get_steps(self):

        # Return as a flat list in order
        params = []
        for piece_name in self.fixed_piece_order:

            piece = getattr(self, piece_name)
            if piece is None:
                continue

            params += conv_to_list(piece)

        return params

    def _get_params_by_step(self):

        return [conv_to_list(getattr(self, piece_name))
                for piece_name in self.fixed_piece_order]

    def _get_indent(self, indent):

        if indent is None:
            return ''
        else:
            ind = 0
            for s in self._p_stack:
                if s == ']':
                    ind += 1
                elif s == '])':
                    ind += 8

            return ' ' * ind

    def _params_print(self, params, indent, _print=print, end='\n'):

        if not isinstance(params, list):
            _print(self._get_indent(indent), params, sep='', end=end)
            return

        elif len(params) == 1:
            self._params_print(params[0], 0, _print=_print)
            return

        elif not isinstance(params, BPtInputMixIn):

            _print(self._get_indent(indent), '[', sep='', end='')
            self._p_stack.append(']')
            self._params_print(params[0], None, _print=_print, end='')
            _print(',', sep='')

        elif isinstance(params, Select):
            _print(self._get_indent(indent), 'Select([', sep='', end='')
            self._p_stack.append('])')
            self._params_print(params[0], None, _print=_print, end='')
            _print(',', sep='')

        for param in params[1:-1]:
            self._params_print(param, 0, _print=_print, end='')
            _print(',', sep='')

        self._params_print(params[-1], 0, _print=_print, end='')
        _print(self._p_stack.pop(), end=end)

        return

    def print_all(self, _print=print):
        '''This method can be used to print a formatted
        representation of this object.'''

        self._p_stack = []

        _print('ModelPipeline')
        _print('-------------')

        pipeline_params = self._get_params_by_step()
        for name, params in zip(self.fixed_piece_order, pipeline_params):

            if params is not None:
                _print(name + '=\\')
                self._params_print(params, 0, _print=_print)
                _print()

        _print('param_search=\\')
        _print(self.param_search)
        _print()


class ProblemSpec(Params):
    '''Problem Spec is defined as an object encapsulating the set of
    parameters used by different :ref:`Evaluation Functions<api.evaluate>`

    Parameters
    ----------
    target : int or str, optional
        The target variable to predict, where the target variable
        is a loaded column within the :class:`Dataset` eventually used
        that has been set to :ref:`Role` target.

        This parameter can be passed either as the name of the
        column, or by passing an integer index. If passed
        an interger index (default = 0), then all loaded target
        variables will be sorted in alphabetical order and that
        index used to select the target to model.

        ::

            default = 0

    scorer : str or list, optional
        Indicator str for which scorer(s) to use when calculating
        validation scores in the context of different
        :ref:`Evaluation Functions<api.evaluate>`.

        A list of str's can be passed as well, in this case, scores for
        all of the requested scorers will be calculated and returned.
        In some cases though, for example :func:`cross_val_score` only
        one scorer can be used, and if passed a list here, the first
        element of the list will be used.

        Note: If using a nested :class:`ParamSearch`, this object
        has its own separate scorer param.

        For a full list of the base sklearn supported scorers please view the
        scikit-learn docs at:
        https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules

        You can also view the BPt reference to these options at :ref:`Scorers`.

        If left as 'default', reasonable scorers will be assigned based
        on the underlying problem type.

        - 'regression'  : ['explained_variance', 'neg_mean_squared_error']
        - 'binary'      : ['matthews', 'roc_auc', 'balanced_accuracy']
        - 'categorical' : ['matthews', 'roc_auc_ovr', 'balanced_accuracy']

        ::

            default = 'default'

    scope : :ref:`Scope`, optional
        This parameter allows for specifying that
        only subset of columns be used in what modelling
        this ProblemSpec is passed to.

        See :ref:`Scope` for a more detailed explained / guide
        on how scopes are defined and used within BPt.

        ::

            default = 'all'

    subjects : :ref:`Subjects`, optional
        This parameter allows for specifying that
        the current experiment be run with only a subset of
        the current subjects.

        A common use of this parameter is to
        specify the reserved keyword 'train' to
        specify that only the training subjects should be used.

        If set to 'default', special behavior will be
        used where if a train/test split is defined then subjects
        will be set to 'train' by default (unless cv='test', then subjects
        will be set to 'all'). If a train/test split is not defined,
        then subjects will be set to 'all'.

        See :ref:`Subjects` for more information
        of the different accepted BPt subject style inputs.

        ::

            default = 'default'

    problem_type : str or 'default', optional
        This parameter controls what type of machine learning
        should be conducted. As either a regression, or classification
        where 'categorical' represents a special case of
        binary classification, where typically a binary
        classifier is trained on each class.

        - 'default'
            Determine the problem type based on how
            the requested target variable is loaded.

        - 'regression', 'f' or 'float'
            For ML on float/continuous target data.

        - 'binary' or 'b'
            For ML on binary target data.

        - 'categorical' or 'c'
            For ML on categorical target data, as multiclass.

        This can almost always be left as default.

        ::

            default = 'default'

    n_jobs : int
        n_jobs are employed within the context
        of a call to Evaluate or Test.

        In general, the way n_jobs are propagated to the different pipeline
        pieces on the backend is that,
        if there is a parameter search, the base
        ML pipeline will all be set to use 1 job,
        and the n_jobs budget will be used
        to train pipelines in parallel to explore different params.
        Otherwise, if no param search,
        n_jobs will be used for each piece individually,
        though some might not support it.

        ::

            default = 1

    random_state : int, RandomState instance or None, optional
        Random state, either as int for a specific seed, or if None then
        the random seed is set by np.random.

        This parameter is used to ensure replicability
        of experiments (wherever possible!).
        In some cases even with a random seed, depending on
        the pipeline pieces being used,
        if any have a component that occassionally yields
        different results, even with the same
        random seed, e.g., some model optimizations,
        then you might still not get
        exact replicability.

        .. note::

            There are some good arguments to be made for not using a
            fixed seed in some cases. See:
            https://scikit-learn.org/stable/common_pitfalls.html#controlling-randomness


        ::

            default = 1

    base_dtype : numpy dtype
        The dataset is cast to a numpy array of float.
        This parameter can be used to change the default
        behavior, e.g., if more resolution or less is needed.

        ::

            default = 'float32'

    '''
    def __init__(self, target=0, scorer='default',
                 scope='all', subjects='default',
                 problem_type='default',
                 n_jobs=1, random_state=1,
                 base_dtype='float32'):

        self.problem_type = problem_type
        self.target = target
        self.scorer = scorer
        self.scope = scope
        self.subjects = subjects
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.base_dtype = base_dtype

        self._checked = False

        self._proc_checks()

    def _proc_checks(self):
        proc_all(self)

    def print_all(self, _print=print):
        '''This method can be used to print a formatted
        representation of this object.'''

        _print('ProblemSpec')
        _print('------------')
        _print('problem_type =', self.problem_type)
        _print('target =', self.target)
        _print('scorer =', self.scorer)
        _print('scope =', self.scope)

        if isinstance(self.subjects, ValueSubset):
            _print('subjects =', self.subjects)
        elif len(self.subjects) < 50:
            _print('subjects =', self.subjects)

        _print('n_jobs =', self.n_jobs)
        _print('random_state =', self.random_state)
        _print()

    def _get_spec(self):

        return {'n_jobs': self.n_jobs,
                'random_state': self.random_state,
                'problem_type': self.problem_type,
                'scope': self.scope}


class CVStrategy(Params):
    '''This objects is used to encapsulate a set of parameters
    representing a cross-validation strategy.

    Parameters
    -----------
    groups : str or None, optional
        The str should refer to the column key in which
        to preserve groups by during any CV splits.
        To create a combination of unique values, use
        :func:`Dataset.add_unique_overlap`.

        Note: the passed column must be of type category as well.

        ::

            default = None

    stratify : str or None, optional
        The str input should refer to a loaded non input
        variable which if of type category. It will assign it
        as a value to preserve
        distribution of groups by during any during any CV splits.

        To create a combination of unique values, use
        :func:`Dataset.add_unique_overlap`.

        Note: the passed column must be of type category as well.

        Any target_cols passed must be categorical
        or binary, and cannot be
        float. Though you may consider creating a binary / k bin copy
        of a float / cont. type target variable.

        ::

            default = None

    train_only_subjects : None or :ref:`subjects`, optional
        If passed  any valid :ref:`subjects` style input here,
        these subjects will be condiered train only and will
        be assigned to every training fold, and never to
        a testing or validation fold.

        ::

            default = None
    '''

    # @TODO add support for test_only ?
    def __init__(self, groups=None, stratify=None,
                 train_only_subjects=None):

        self.groups = groups
        self.stratify = stratify
        self.train_only_subjects = train_only_subjects

        if groups is not None and stratify is not None:
            raise RuntimeError('Warning: BPt does not currently '
                               'support groups and'
                               'stratify together!'
                               'Please use only one.')

        if not isinstance(groups, str) and groups is not None:
            raise RuntimeError('Passing non str for groups is depreciated. '
                               ' Instead, first use the function'
                               ' add_unique_overlap(cols=..., new_cols=...) to'
                               ' create explicitly the specified new values.')

        if not isinstance(stratify, str) and stratify is not None:
            raise RuntimeError('Passing non str for stratify is depreciated. '
                               ' Instead, first use the function'
                               ' add_unique_overlap(cols=..., new_cols=...) to'
                               ' create explicitly the specified new values.')


class CV(Params):
    ''' This object is used to define a BPt style custom
    CV strategy, e.g., as KFold

    Parameters
    ----------
    splits : int, float, str or list of str, optional
        `splits` allows you to specify the base of what CV strategy
        should be used.

        Specifically, options for split are:

        - int
            The number of k-fold splits to conduct. (E.g., 3 for
            3-fold CV split to be conducted at
            every hyper-param evaluation).

        - float
            Must be 0 < `splits` < 1, and defines a
            single train-test like split,
            with `splits` % of the current training data size used
            as a validation set.

        - str
            If a str is passed, then it must correspond to a
            loaded categorical non input variable. In
            this case, a leave-out-group CV will be used according
            to the value of the variable.

        Note that the parameter n_repeats is designed to work with
        any of these choices.

        ::

            default = 3

    n_repeats : int, optional
        The number of times to repeat the defined strategy
        as defined in `splits`.

        For example, if `n_repeats` is set to 2, and `splits` is 3,
        then a twice repeated 3-fold CV will be performed

        ::

            default = 1

    cv_strategy : None or :class:`CVStrategy`, optional
        Optional cv_strategy to employ for calculating splits.
        If passed None, use no strategy.

        See :class:`CVStrategy`

        Can also pass valid cv_strategy args seperately.
        If any passed, they will override any values
        set in the pass cv_strategy if any.

        ::

            default = None

    random_state : 'context', int or None, optional
        The fixed random seed in which this
        CV object should adhere to.

        If left as default value of 'context', then
        the random state will be set based on the
        context of where it is called, i.e., typically
        the random_state set in :class:`ProblemSpec`.

        ::

            default = 'context'

    only_fold : int, list of int, or None, optional
        This parameter specifies if special subset of the requested
        CV folds should be used. If kept as None, normal all
        fold behavior will be used. Otherwise, if passed as an int,
        then that int must represent a valid cv fold, e.g.,

        ::

            only_fold = 0

        Would run the CV but only the 1st fold. Likewise if
        a list is passed,

        ::

            only_fold = [0, 2]

        Then only the first and 3rd folds will be run.
        This parameter is useful in cases where the base
        experiment is too computationally intensive, and
        it is desired to run a complete CV but in smaller
        chunks.

        .. warning::

            When used with n_repeats > 1, only_fold
            will index folds from the repeats, e.g.,
            split=2, n_repeats=2, only_fold can be 0, 1, 2, or 3,
            but functionally for say computing summary scores, like std across
            repeats, n_repeats will be treated as 1.

            For example if passed with the setup above only_fold=[0, 1, 2],
            then progress bars and summary stats will still show n_repeats=1.

        ::

            default = None

    cv_strategy_kwargs : kwargs, optional
        If any additional parameters are passed in kwargs style,
        e.g., ::

            splits = 3

        Then they will try to be set in the base cv_strategy.

    '''

    def __init__(self, splits=3, n_repeats=1,
                 cv_strategy=None, random_state='context',
                 only_fold=None,
                 **cv_strategy_kwargs):

        self.splits = splits
        self.n_repeats = n_repeats
        self.cv_strategy = cv_strategy
        self.random_state = random_state
        self.only_fold = only_fold

        if self.cv_strategy is None:
            self.cv_strategy = CVStrategy()

        self.cv_strategy.set_params(**cv_strategy_kwargs)

    def _apply_dataset(self, dataset):
        return get_bpt_cv(self, dataset)

# Depreciations
#############################


def depreciate(func):

    # Get current and new names
    name = str(func.__name__)
    new_name = str(func.new_class.__name__)

    warn_message = name + ' is depreciated and in a future version will'
    warn_message += ' be removed. Use the new class '
    warn_message += new_name + ' to mute this warning.'

    class DepreciatedClass():

        __doc__ = warn_message

        def __new__(cls, *args, **kwargs):

            # Warn on creation
            warnings.warn(warn_message)

            # Return new class instead of this one
            return func.new_class(*args, **kwargs)

    return DepreciatedClass


@depreciate
class Feat_Selector():
    new_class = FeatSelector


@depreciate
class Param_Search():
    new_class = ParamSearch


@depreciate
class Model_Pipeline():
    new_class = ModelPipeline


@depreciate
class Problem_Spec():
    new_class = ProblemSpec



import numpy as np
from itertools import combinations
from .helpers import proc_fop
from pandas.util._decorators import doc
from .Dataset import _shared_docs, _sip_docs, _file_docs


def _drop_subjects(self, subjects):

    if len(subjects) > 0:
        self.drop(subjects, axis=0, inplace=True)
        self._print('Dropped', len(subjects), 'Rows', level=1)
        self._print('Dropped Rows:', subjects, level=2)


@doc(**_sip_docs)
def drop_nan_subjects(self, scope, inplace=False):
    '''This method is used for
    dropping all of the subjects which have NaN
    values for a given scope / column.

    Parameters
    ----------
    {scope}

    {inplace}

    Examples
    ---------

    .. ipython:: python

        data = bp.read_csv('data/example1.csv')
        data
        data.drop_nan_subjects(scope='all')
        data.drop_nan_subjects(scope='numbers')

    '''

    if not inplace:
        return self._inplace('drop_nan_subjects', locals())

    cols = self.get_cols(scope)

    if len(cols) == 0:
        raise RuntimeError(f'No columns found for scope: {scope}.')

    nan_subjects = self._get_nan_subjects(cols[0])
    for col in cols[1:]:
        nan_subjects = nan_subjects.union(self._get_nan_subjects(col))

    # Drop all nan subjects
    self._drop_subjects(nan_subjects)


_si_docs = {'subjects': _shared_docs['subjects'],
            'inplace': _shared_docs['inplace']}


@doc(**_si_docs)
def apply_inclusions(self, subjects, inplace=False):
    '''This method will drop all subjects
    that do not overlap with the passed subjects to
    this function. In this sense, this method acts
    like a whitelist, where you could pass for example
    only valid subjects that passed some QC, and all
    other's loaded will be dropped.

    Parameters
    -----------
    {subjects}

    {inplace}

    '''

    if not inplace:
        return self._inplace('apply_inclusions', locals())

    # Load inclusions
    inclusions = self.get_subjects(subjects, return_as='set')

    if len(inclusions) > 0:
        to_drop = set(self.index) - inclusions
        self.drop(list(to_drop), axis=0, inplace=True)


@doc(**_si_docs)
def apply_exclusions(self, subjects, inplace=False):
    '''This method will drop all subjects
    that overlap with the passed subjects to
    this function.

    Parameters
    -----------
    {subjects}

    {inplace}

    '''

    if not inplace:
        return self._inplace('apply_exclusions', locals())

    # Load exclusions
    exclusions = self.get_subjects(subjects, return_as='set')

    if len(exclusions) > 0:
        to_drop = set(self.index).intersection(exclusions)
        self.drop(list(to_drop), axis=0, inplace=True)


@doc(**_sip_docs)
def drop_subjects_by_nan(self, scope='all', threshold=.5, inplace=False):
    '''This method is used for dropping subjects based on
    the amount of missing values found across a subset of
    columns as selected by scope. Each subject is dropped
    if it has greater than or equal to the passed threshold
    of NaN values.

    Parameters
    -----------
    {scope}
        ::

            default = 'all'

    threshold : float or int, optional
        Passed as a float between 0 and 1, or
        as an int. If greater than 0 or less than 1,
        this threshold represents the percent of columns
        a subject needs to have missing for it to be dropped.
        If 1 or over, then it represents an absolute number of
        columns that a subject needs to have greater than or equal
        to that value in order to drop.

        ::

            default = .5

    {inplace}

    '''

    if not inplace:
        return self._inplace('drop_subjects_by_nan', locals())

    # Get cols from scope
    cols = self.get_cols(scope)

    # Get nan counts by column
    col_nan_cnts = self[cols].isnull().sum(axis=1)

    # If between 0 and 1
    if threshold > 0 and threshold < 1:

        # Change threshold from percent to abs
        threshold = threshold * len(cols)
        self._print('Setting NaN threshold to:', threshold, level=1)

    # Calculate subjects to drop if greater than or equal to threshold
    to_drop = self.loc[col_nan_cnts >= threshold].index

    self._drop_subjects(to_drop)


def _drop_or_nan(self, col, to_drop_index, all_to_drop, drop):
    '''Internal helper function for commonly re-used drop or
    nan function.'''

    # If drop, add to drop list at end
    if drop:
        all_to_drop.update(set(to_drop_index))

    # Otherwise, set to NaN in place
    else:
        self.loc[to_drop_index, col] = np.nan


def _data_file_fail_check(self, cols):

    for col in cols:
        if 'data file' in self.scopes[col]:
            raise RuntimeError('Loaded column: ' + col + ' cannot be a'
                               ' data file.')


def _drop_cols(self, to_drop):

    if len(to_drop) > 0:

        self.drop(to_drop, axis=1, inplace=True)
        self._print('Dropped', len(to_drop), 'Columns', level=1)
        self._print('Dropped Columns:', to_drop, level=2)


_outlier_docs = {**_sip_docs, **_file_docs, 'drop': _shared_docs['drop']}


@doc(**_outlier_docs)
def filter_outliers_by_percent(self, scope='float', fop=1, drop=True,
                               reduce_func=np.mean, n_jobs=-1, inplace=False):
    '''This method is designed to allow dropping a fixed percent of outliers
    from the requested columns. This method is designed to work
    on float type / cont. variables.

    Note: This method operates on each of the columns specified by scope
    independently. In the case that multiple columns are passed, then the
    overlap of all outliers from each column will dropped after all
    have been calculated (therefore the order won't matter).

    This method can be used with data file's as well, the
    reduce_func and n_jobs parameters are specific to this case.

    Parameters
    -----------
    {scope}
        ::

            default = 'float'

    fop : float, tuple, optional
        This parameter represents the percent of outliers to drop.
        It should be passed as a percent, e.g., therefore 1 for
        one percent, or 5 for five percent.

        This can also be passed as a tuple with two elements, where
        the first entry represents the percent to filter from the lower
        part of the distribution and the second element the percent from
        the upper half of the distribution. For example,

        ::

            filter_outlier_percent = (5, 1)

        This set of parameters with drop 5 percent from the lower part
        of the distribution and only 1 percent from the top portion.
        Likewise, you can use None on one side to skip dropping from
        one half, for example:

        ::

            filter_outlier_percent = (5, None)

        Would drop only five percent from the bottom half, and not drop
        any from the top half.

        ::

            default = 1

    {drop}

    {reduce_func}

    {n_jobs}

    {inplace}

    '''

    if not inplace:
        return self._inplace('filter_outliers_by_percent', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # For if to drop
    all_to_drop = set()

    # Proc fop
    fop = proc_fop(fop)
    lower, upper = fop

    for col in cols:

        # Extract non-nan values
        values = self._get_values(col, dropna=True,
                                  reduce_func=reduce_func,
                                  n_jobs=n_jobs)

        if lower is not None:

            # Compute index under lower bound
            mask = values < values.quantile(lower)
            to_drop_index = values[mask].index

            # Drop or NaN
            self._drop_or_nan(col=col, to_drop_index=to_drop_index,
                              all_to_drop=all_to_drop, drop=drop)

        if upper is not None:

            # Compute index above upper bound
            mask = values > values.quantile(upper)
            to_drop_index = values[mask].index

            # Drop or NaN
            self._drop_or_nan(col=col, to_drop_index=to_drop_index,
                              all_to_drop=all_to_drop, drop=drop)

    # Drop now, if to drop
    if drop:
        self.drop(list(all_to_drop), axis=0, inplace=True)

    # Update file mapping if needed
    self._check_file_mapping()


@doc(**_outlier_docs)
def filter_outliers_by_std(self, scope='float', n_std=10, drop=True,
                           reduce_func=np.mean, n_jobs=-1, inplace=False):
    '''This method is designed to allow dropping outliers
    from the requested columns based on comparisons with that columns
    standard deviation.

    Note: This method operates on each of the columns specified by scope
    independently. In the case that multiple columns are passed, then the
    overlap of all outliers from each column will dropped after all
    have been calculated (therefore the order won't matter).

    This method can be used with data file's as well, the
    reduce_func and n_jobs parameters are specific to this case.

    Parameters
    -----------
    {scope}
        ::

            default = 'float'

    n_std : float, tuple, optional
        This value is used to set an outlier threshold by
        standrad deviation. For example if passed n_std = 10,
        then it will be converted internally to (10, 10).
        This parameter determines outliers as
        data points within each
        relevant column (as determined by the scope argument) where their
        value is less than the mean of the
        column - n_std[0] * the standard deviation of the column,
        and greater than the mean of the column + n_std[1] * the standard
        deviation of the column.

        If a single number is passed, that number is applied to
        both the lower
        and upper range.
        If a tuple with None on one side is passed, e.g.
        (None, 3), then nothing will be taken off
        that lower or upper bound.

        ::

            default = 10

    {drop}

    {reduce_func}

    {n_jobs}

    {inplace}

    Examples
    ----------

    If we define a dataset, we can check the std.

    .. ipython:: python

        import BPt as bp
        import numpy as np
        data = bp.Dataset()
        data.verbose = 1
        data['1'] = [1, 1, 2, 2, 3, 3, 4, 4, 5, 5]
        np.std(data['1']), np.mean(data['1'])

    We can now see how different thresholds work.

    .. ipython:: python

        # This won't drop anything
        data.filter_outliers_by_std(n_std=2)

        # This will
        data.filter_outliers_by_std(n_std=1)

    What if there was more than one column?

    .. ipython:: python

        data['2'] = [1, 1, 1, 1, 10, 1, 1, 1, 1, 1]

        # Now a subject will be dropped
        data.filter_outliers_by_std(n_std=2)
        data.filter_outliers_by_std(n_std=1)

    We can also apply it only to one column, and instead of dropping
    subjects, replace outliers with NaN's

    .. ipython:: python

        data.filter_outliers_by_std(n_std=1, scope='1', drop=False)

    '''

    if not inplace:
        return self._inplace('filter_outliers_by_std', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # For if to drop
    all_to_drop = set()

    # Wrap correctly passed n_std
    if not isinstance(n_std, tuple):
        n_std = (n_std, n_std)

    for col in cols:

        # Extract non-nan values
        values = self._get_values(col, dropna=True,
                                  reduce_func=reduce_func,
                                  n_jobs=n_jobs)

        mean = values.mean()
        std = values.std()

        # If lower
        if n_std[0] is not None:
            l_scale = mean - (n_std[0] * std)
            to_drop_index = values[values < l_scale].index

            # Drop or NaN
            self._drop_or_nan(col=col, to_drop_index=to_drop_index,
                              all_to_drop=all_to_drop, drop=drop)

        # If upper
        if n_std[1] is not None:
            u_scale = mean + (n_std[1] * std)
            to_drop_index = values[values > u_scale].index

            # Drop or NaN
            self._drop_or_nan(col=col, to_drop_index=to_drop_index,
                              all_to_drop=all_to_drop, drop=drop)

    # Drop now, if to drop
    if drop:
        self._drop_subjects(all_to_drop)

    # Update file mapping if needed
    self._check_file_mapping()


@doc(**_sip_docs, drop=_shared_docs['drop'])
def filter_categorical_by_percent(self, scope='category', drop_percent=1,
                                  drop=True, inplace=False):
    '''This method is designed to allow performing outlier filtering
    on categorical type variables. Note that this method assume
    all columns passed are of type 'category', and they if not already
    will be cast first to pandas data type 'category'.

    Note: NaN values will be skipped. If desired to treat them as a class,
    use the method nan_to_class to first. It is worth noting further that
    this method will not work on data files.

    This method operates on each of the columns specified by scope
    independently. In the case that multiple columns are passed, then the
    overlap of all outliers from each column will dropped after all
    have been calculated (therefore the order won't matter).

    Parameters
    -----------
    {scope}
        ::

            default = 'category'

    drop_percent : float, optional
        This parameter acts as a percentage threshold for dropping
        categories when loading categorical data. This parameter
        should be passed as a percent, such that a category
        will be dropped if it makes up less than that % of the data points.
        For example:

        ::

            drop_percent = 1

        In this case any data points within the
        relevant categories as specified by scope
        with a category constituting less than 1% of total
        valid data points will be dropped (or set to NaN if drop=False).

    {drop}

    {inplace}

    '''

    if not inplace:
        return self._inplace('filter_categorical_by_percent', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # Check for data files
    self._data_file_fail_check(cols)

    # For if to drop
    all_to_drop = set()

    # Divide drop percent by 100
    dp = drop_percent / 100

    for col in cols:

        # Make sure categorical
        self._add_scope(col, 'category')

        # Extract non-nan values
        values = self._get_values(col, dropna=True)

        # Get to drop subjects
        unique_vals, counts = np.unique(values, return_counts=True)
        drop_inds = np.where((counts / len(values)) < dp)

        drop_vals = unique_vals[drop_inds]
        to_drop_index = values.index[values.isin(drop_vals)]

        # Drop or NaN
        self._drop_or_nan(col=col, to_drop_index=to_drop_index,
                          all_to_drop=all_to_drop, drop=drop)

    # Drop now, if to drop
    if drop:
        self._drop_subjects(all_to_drop)


@doc(**_sip_docs)
def drop_id_cols(self, scope='all', inplace=False):
    '''This method will drop any str-type / object type columns
    where the number of unique columns is equal
    to the length of the dataframe.

    Parameters
    ----------
    {scope}
        ::

            default = 'all'

    {inplace}

    '''

    if not inplace:
        return self._inplace('drop_id_cols', locals())

    # Get cols from scope
    cols = self.get_cols(scope)

    to_drop = []

    for col in cols:
        if self[col].dtype.name == 'object':
            if len(self[col].unique()) == len(self):
                to_drop.append(col)

    self._drop_cols(to_drop)


@doc(**_sip_docs)
def drop_duplicate_cols(self, scope='all', inplace=False):
    '''This method is used for checking to see if there are
    any columns loaded with duplicate values. If there is, then
    one of the duplicates will be dropped.

    {scope}
        ::

            default = 'all'

    {inplace}

    '''

    if not inplace:
        return self._inplace('drop_duplicate_cols', locals())

    # Get cols from scope
    cols = self.get_cols(scope)

    to_drop = []
    for col1, col2 in combinations(cols, 2):
        if np.array_equal(self[col1], self[col2]):
            to_drop.append(col2)

    self._drop_cols(to_drop)


@doc(**_sip_docs)
def drop_cols(self,
              scope='all',
              exclusions=None,
              inclusions=None,
              inplace=False):
    '''This method is designed to allow
    dropping columns based on some flexible arguments.
    Essentially, exclusions, inclusions and scope are
    all scope style arguments for selecting subsets of columns,
    and the way they are composed to specify dropping a column
    is as follows:

    For any given column, if within the columns selected
    by scope and EITHER in the passed exclusions columns or
    not in the passed inclusions columns, drop that column.

    Parameters
    ------------
    {scope}
        ::

            default = 'all'

    exclusions : :ref:`Scope`, optional
        A BPt style :ref:`Scope` used to select a subset of
        columns in which if a column is in the selected
        subset of columns, it will be dropped (though if scope is
        set, then only those columns within scope will be checked to
        see if inside of the passed inclusions here.)


    inclusions : :ref:`Scope`, optional
        A BPt style :ref:`Scope` used to select a subset of
        columns in which if a column is not in the selected
        subset of columns, it will be dropped (though if scope is
        set, then only those columns within scope will be checked to
        see if outside of the passed inclusions here.)

    {inplace}

    Examples
    ---------

    Given the below dataset, we could try dropping some different columns.
    Each drop will be done on a copy of the Dataset.

    .. ipython:: python

        import BPt as bp
        data = bp.Dataset(columns=['cat1', 'cat2', 'cat3', 'dog1', 'dog2'])
        list(data)

        list(data.drop_cols(exclusions='cat'))
        list(data.drop_cols(exclusions='2', scope='cat'))
        list(data.drop_cols(inclusions=['1', '2'], scope='cat'))
        list(data.drop_cols(inclusions='dog'))

    '''

    if not inplace:
        return self._inplace('drop_cols', locals())

    # Get cols from scope
    cols = self.get_cols(scope)

    if exclusions is not None:
        exclusions = self.get_cols(exclusions)
    else:
        exclusions = []

    if inclusions is not None:
        inclusions = self.get_cols(inclusions)
    else:
        inclusions = cols

    to_drop = []
    for col in cols:
        if col in exclusions or col not in inclusions:
            to_drop.append(col)

    self._drop_cols(to_drop)


@doc(**_sip_docs)
def drop_cols_by_unique_val(self, scope='all', threshold=1,
                            dropna=True, inplace=False):
    '''This method will drop any columns with less than or equal to
    the number of unique values.
    This is a coarse filtering method for removing
    uninformative variables which may have been
    loaded or included with a dataset.

    Parameters
    -----------
    {scope}
        ::

            default = 'all'

    threshold : int, optional
        The threshold in which a column should be dropped
        if it has less than or equal to this number of
        unique values.

        For example, the default value of 1, will
        only drop column with only 1 unique value.

        ::

            default = 1

    dropna : bool, optional
        This parameter controls if NaN values are
        ignored when determining how many unique values there
        are, when set to True, and if set to False will
        count NaN as a unique value.

        ::

            default = True

    {inplace}
    '''

    if not inplace:
        return self._inplace('drop_cols_by_unique_val', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    to_drop = []
    for col in cols:
        values = self._get_values(col, dropna=dropna)
        if len(values.unique()) <= threshold:
            to_drop.append(col)

    self._drop_cols(to_drop)


@doc(**_sip_docs)
def drop_cols_by_nan(self, scope='all', threshold=.5, inplace=False):
    '''This method is used for dropping columns based on
    the amount of missing values found across all subjects.
    Each column is dropped if it has greater than or equal
    to the passed threshold of NaN values.

    Parameters
    -----------
    {scope}
        ::

            default = 'all'

    threshold : float or int, optional
        Passed as a float between 0 and 1, or
        as an int. If greater than 0 or less than 1,
        this parameter represents the threshold in which
        a column should be dropped if it has greater than or
        equal to this percent of its columns as NaN values.

        If greater than 1, then this threshold represents the
        absolute value in which if a column has that number
        of subjects or greater with NaN, it will be dropped.

        For example, if a column has 3 values and 7 NaN values,
        passing .7 or lower here would drop this column, but passing
        anything above .7, e.g., .8 would not.

        ::

            default = .5

    {inplace}

    '''

    if not inplace:
        return self._inplace('drop_cols_by_nan', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # Change threshold from percent to abs
    if threshold > 0 and threshold < 1:
        threshold = threshold * self.shape[0]
        self._print('Setting NaN threshold to:', threshold, level=1)

    to_drop = []
    for col in cols:
        values = self._get_values(col, dropna=False)
        nan_percent = values.isnull().sum()

        if nan_percent >= threshold:
            to_drop.append(col)

    self._drop_cols(to_drop)



from ..helpers import get_obj_and_params, all_from_objects
from ...extensions.loaders import Identity, SurfLabels

LOADERS = {
    'identity': (Identity, ['default']),
    'surface rois': (SurfLabels, ['default']),
}

# If nilearn dependencies
try:
    from nilearn.input_data import NiftiLabelsMasker
    from ...extensions.loaders import SingleConnectivityMeasure
    LOADERS['volume rois'] = (NiftiLabelsMasker, ['default'])
    LOADERS['connectivity'] = (SingleConnectivityMeasure, ['default'])

except ImportError:
    pass


def get_loader_and_params(loader_str, extra_params, params, **kwargs):

    loader, extra_loader_params, loader_params =\
        get_obj_and_params(loader_str, LOADERS, extra_params, params)

    return loader(**extra_loader_params), loader_params


all_obj_keys = all_from_objects(LOADERS)
